from collections import OrderedDict
from collections.abc import Iterable
from typing import Literal, Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from matplotlib.colors import Normalize

from openretina.data_io.hoefling_2024.dataloaders import get_dims_for_loader_dict
from openretina.modules.core.base_core import Core
from openretina.modules.core.base_core import Core3d
from openretina.modules.layers.convolutions import STSeparableBatchConv3d
from openretina.modules.layers.convolutions import temporal_smoothing
from openretina.modules.layers.convolutions import compute_temporal_kernel
from openretina.modules.layers.scaling import Bias3DLayer, Scale2DLayer, Scale3DLayer
from openretina.modules.layers.convolutions import TorchFullConv3D
from openretina.modules.layers.convolutions import TorchSTSeparableConv3D
from openretina.modules.layers.convolutions import TimeIndependentConv3D
from openretina.modules.layers.regularizers import FlatLaplaceL23dnorm, TimeLaplaceL23dnorm
from openretina.utils.misc import set_seed
from openretina.utils.model_utils import eval_state

class Encoder(nn.Module):
    """
    puts together all parts of model (core, readouts) and defines a forward
    function which will return the output of the model; PyTorch then allows
    to call .backward() on the Encoder which will compute the gradients
    """

    def __init__(
        self,
        core: nn.Module,
        readout,
    ):
        super().__init__()
        self.core = core
        self.readout = readout
        self.detach_core = False

    def forward(self, x: torch.Tensor, data_key: str | None = None, detach_core: bool = False, **kwargs):
        self.detach_core = detach_core
        x = self.core(x, data_key=data_key)
        if self.detach_core:
            x = x.detach()
        x = self.readout(x, data_key=data_key)
        return x

    def readout_keys(self) -> list[str]:
        return self.readout.readout_keys()


def get_module_output_shape(
    model: torch.nn.Module, input_shape: tuple[int, ...], use_cuda: bool = True
) -> tuple[int, ...]:
    """
    Return the output shape of the model when fed in an array of `input_shape`.
    Note that a zero array of shape `input_shape` is fed into the model and the
    shape of the output of the model is returned.

    Args:
        model (nn.Module): PyTorch module for which to compute the output shape
        input_shape (tuple): Shape specification for the input array into the model
        use_cuda (bool, optional): If True, model will be evaluated on CUDA if available. Othewrise
            model evaluation will take place on CPU. Defaults to True.

    Returns:
        tuple: output shape of the model

    """
    # infer the original device
    initial_device = next(iter(model.parameters())).device
    device = "cuda" if torch.cuda.is_available() and use_cuda else "cpu"
    with eval_state(model):
        with torch.no_grad():
            inp_ = torch.zeros(1, *input_shape[1:], device=device)
            output = model.to(device)(inp_)
    model.to(initial_device)
    return output.shape


class ParametricFactorizedBatchConv3dCore(Core3d):
    def __init__(
        self,
        n_neurons_dict,
        input_channels: int = 2,
        num_scans: int = 1,
        hidden_channels=[8],
        temporal_kernel_size=[21],
        spatial_kernel_size=[14],
        layers: int = 1,
        gamma_hidden: float = 0.0,
        gamma_input: float = 0.0,
        gamma_in_sparse: float = 0.0,
        gamma_temporal: float = 0.0,
        final_nonlinearity: bool = True,
        bias: bool = True,
        momentum: float = 0.1,
        input_padding: bool = False,
        hidden_padding: bool = True,
        batch_norm: bool = True,
        batch_norm_scale: bool = True,
        laplace_padding: int = 0,
        stack=None,
        batch_adaptation: bool = True,
        use_avg_reg: bool = False,
        nonlinearity: str = "ELU",
        conv_type: str = "custom_separable",
    ):
        super().__init__()
        self._input_weights_regularizer_spatial = FlatLaplaceL23dnorm(padding=laplace_padding)
        self._input_weights_regularizer_temporal = TimeLaplaceL23dnorm(padding=laplace_padding)

        if conv_type == "separable":
            self.conv_class: type[torch.nn.Module] = TorchSTSeparableConv3D
        elif conv_type == "custom_separable":
            self.conv_class = STSeparableBatchConv3d
        elif conv_type == "full":
            self.conv_class = TorchFullConv3D
        elif conv_type == "time_independent":
            self.conv_class = TimeIndependentConv3D
        else:
            raise ValueError(f"Un-implemented conv_type {conv_type}")

        self.num_scans = num_scans
        self.layers = layers
        self.gamma_input = gamma_input
        self.gamma_in_sparse = gamma_in_sparse
        self.gamma_hidden = gamma_hidden
        self.gamma_temporal = gamma_temporal
        self.input_channels = input_channels
        self.hidden_channels = hidden_channels
        self.use_avg_reg = use_avg_reg

        self.features = nn.Sequential()
        if stack is None:
            self.stack = list(range(self.layers))
        else:
            self.stack = [range(self.layers)[stack]] if isinstance(stack, int) else stack  # type: ignore

        log_speed_dict = {}
        for k in n_neurons_dict:
            var_name = "_".join(["log_speed", str(k)])
            log_speed_val = torch.nn.Parameter(data=torch.zeros(1), requires_grad=batch_adaptation)
            setattr(self, var_name, log_speed_val)
            log_speed_dict[var_name] = log_speed_val

        if input_padding:
            input_pad: tuple[int, int, int] | int = (0, spatial_kernel_size[0] // 2, spatial_kernel_size[0] // 2)
        else:
            input_pad = 0
        if hidden_padding & (len(spatial_kernel_size) > 1):
            hidden_pad: list[tuple[int, int, int] | int] = [
                (0, spatial_kernel_size[x] // 2, spatial_kernel_size[x] // 2)
                for x in range(1, len(spatial_kernel_size))
            ]
        else:
            hidden_pad = [0 for _ in range(1, len(spatial_kernel_size))]

        if not isinstance(hidden_channels, Iterable):
            hidden_channels = [hidden_channels] * self.layers
        if not isinstance(temporal_kernel_size, Iterable):
            temporal_kernel_size = [temporal_kernel_size] * self.layers
        if not isinstance(spatial_kernel_size, Iterable):
            spatial_kernel_size = [spatial_kernel_size] * self.layers

        # --- first layer
        layer = OrderedDict()
        layer["conv"] = self.conv_class(
            input_channels,
            hidden_channels[0],
            log_speed_dict,
            temporal_kernel_size[0],
            spatial_kernel_size[0],
            bias=False,
            padding=input_pad,
            num_scans=self.num_scans,
        )
        if batch_norm:
            layer["norm"] = nn.BatchNorm3d(
                hidden_channels[0], momentum=momentum, affine=bias and batch_norm_scale
            )  # ok or should we ensure same batch norm?
            if bias:
                if not batch_norm_scale:
                    layer["bias"] = Bias3DLayer(hidden_channels[0])
            elif batch_norm_scale:
                layer["scale"] = Scale3DLayer(hidden_channels[0])
        if final_nonlinearity:
            layer["nonlin"] = getattr(nn, nonlinearity)()
        self.features.add_module("layer0", nn.Sequential(layer))

        # --- other layers

        for layer_num in range(1, self.layers):
            layer = OrderedDict()
            layer["conv"] = self.conv_class(
                hidden_channels[layer_num - 1],
                hidden_channels[layer_num],
                log_speed_dict,
                temporal_kernel_size[layer_num],
                spatial_kernel_size[layer_num],
                bias=False,
                padding=hidden_pad[layer_num - 1],
                num_scans=self.num_scans,
            )
            if batch_norm:
                layer["norm"] = nn.BatchNorm3d(
                    hidden_channels[layer_num], momentum=momentum, affine=bias and batch_norm_scale
                )
                if bias:
                    if not batch_norm_scale:
                        layer["bias"] = Bias3DLayer(hidden_channels[layer_num])
                elif batch_norm_scale:
                    layer["scale"] = Scale2DLayer(hidden_channels[layer_num])
            if final_nonlinearity or layer_num < self.layers - 1:
                layer["nonlin"] = getattr(nn, nonlinearity)()
            self.features.add_module("layer{}".format(layer_num), nn.Sequential(layer))

        self.apply(self.init_conv)

    def forward(self, input_, data_key=None):
        ret = []
        do_skip = False
        for layer_num, feat in enumerate(self.features):
            input_ = feat((torch.cat(ret[-min(self.skip, layer_num) :], dim=1) if do_skip else input_, data_key))
            ret.append(input_)

        return torch.cat([ret[ind] for ind in self.stack], dim=1)

    def spatial_laplace(self):
        return self._input_weights_regularizer_spatial(self.features[0].conv.weight_spatial, avg=self.use_avg_reg)

    def group_sparsity(self):  # check if this is really what we want
        ret = 0
        for layer_num in range(1, self.layers):
            ret = (
                ret
                + (
                    self.features[layer_num].conv.weight_spatial.pow(2).sum([2, 3, 4]).sqrt().sum(1)
                    / torch.sqrt(1e-8 + self.features[layer_num].conv.weight_spatial.pow(2).sum([1, 2, 3, 4]))
                ).sum()
            )
        return ret

    def group_sparsity0(self):  # check if this is really what we want
        weight_temporal = compute_temporal_kernel(
            torch.zeros(1, device=self.features[0].conv.sin_weights.device),
            self.features[0].conv.sin_weights,
            self.features[0].conv.cos_weights,
            self.features[0].conv.temporal_kernel_size,
        )
        # abc are dummy dimensions
        weight = torch.einsum("oitab,oichw->oithw", weight_temporal, self.features[0].conv.weight_spatial)
        ret = (weight.pow(2).sum([2, 3, 4]).sqrt().sum(1) / torch.sqrt(1e-8 + weight.pow(2).sum([1, 2, 3, 4]))).sum()
        return ret

    def temporal_smoothness(self):
        ret = 0
        for layer_norm in range(self.layers):
            ret += temporal_smoothing(
                self.features[layer_norm].conv.sin_weights, self.features[layer_norm].conv.cos_weights
            )
        return ret

    def regularizer(self):
        if self.conv_class == STSeparableBatchConv3d:
            return (
                self.group_sparsity() * self.gamma_hidden
                + self.gamma_input * self.spatial_laplace()
                + self.gamma_temporal * self.temporal_smoothness()
                + self.group_sparsity0() * self.gamma_in_sparse
            )
        else:
            return 0

    @property
    def outchannels(self) -> list[int]:
        return len(self.features) * self.hidden_channels[-1]


class SpatialXFeature3d(nn.Module):
    def __init__(
        self,
        in_shape: tuple[int, int, int, int],
        outdims: int,
        gaussian_masks: bool = False,
        gaussian_mean_scale: float = 1e0,
        gaussian_var_scale: float = 1e0,
        initialize_from_roi_masks: bool = False,
        roi_mask=[],
        positive: bool = False,
        scale: bool = False,
        bias: bool = True,
        nonlinearity: bool = True,
    ):
        """
        TODO write docstring

        Args:
            in_shape (tuple): The shape of the input tensor (c, t, w, h).
            outdims (int): The number of output dimensions (usually the number of neurons in the session).
            gaussian_masks (bool, optional): Whether to use Gaussian masks. Defaults to False.
            gaussian_mean_scale (float, optional): The scale factor for the Gaussian mean. Defaults to 1e0.
            gaussian_var_scale (float, optional): The scale factor for the Gaussian variance. Defaults to 1e0.
            initialize_from_roi_masks (bool, optional): Whether to initialize from ROI masks. Defaults to False.
            roi_mask (list, optional): The ROI mask. Defaults to an empty list.
            positive (bool, optional): Whether the output should be positive. Defaults to False.
            scale (bool, optional): Whether to include a scale parameter. Defaults to False.
            bias (bool, optional): Whether to include a bias parameter. Defaults to True.
            nonlinearity (bool, optional): Whether to include a nonlinearity. Defaults to True.
        """
        super().__init__()
        self.in_shape = in_shape
        c, t, w, h = in_shape
        self.outdims = outdims
        self.gaussian_masks = gaussian_masks
        self.gaussian_mean_scale = gaussian_mean_scale
        self.gaussian_var_scale = gaussian_var_scale
        self.positive = positive
        self.nonlinearity = nonlinearity
        self.initialize_from_roi_masks = initialize_from_roi_masks

        if gaussian_masks:
            """we train on the log var and transform to var in a separate step"""
            self.mask_mean = torch.nn.Parameter(data=torch.zeros(self.outdims, 2), requires_grad=True)
            self.mask_log_var = torch.nn.Parameter(data=torch.zeros(self.outdims), requires_grad=True)
            self.grid = torch.nn.Parameter(data=self.make_mask_grid(w, h), requires_grad=False)
            self.masks = self.normal_pdf().permute(1, 2, 0)
        else:
            if initialize_from_roi_masks:
                self.mask_mean = torch.nn.Parameter(data=roi_mask, requires_grad=False)
                self.mask_var = torch.nn.Parameter(data=torch.ones(self.outdims) * 0.01, requires_grad=False)
                self.grid = torch.nn.Parameter(data=self.make_mask_grid(w, h), requires_grad=False)
                self.masks = nn.Parameter(self.get_normal_pdf_from_roi_mask().permute(1, 2, 0))
            else:
                self.masks = nn.Parameter(torch.Tensor(w, h, outdims))

        self.features = nn.Parameter(torch.Tensor(1, c, 1, outdims))

        if scale:
            scale_param = nn.Parameter(torch.Tensor(outdims))
            self.register_parameter("scale", scale_param)
        else:
            self.register_parameter("scale", None)

        if bias:
            bias_param = nn.Parameter(torch.Tensor(outdims))
            self.register_parameter("bias", bias_param)
        else:
            self.register_parameter("bias", None)

        self.initialize()

    def initialize(self) -> None:
        if (not self.gaussian_masks) and (not self.initialize_from_roi_masks):
            self.masks.data.normal_(0.0, 0.01)
        self.features.data.normal_(0.0, 0.01)
        if self.scale is not None:
            self.scale.data.normal_(1.0, 0.01)
        if self.bias is not None:
            self.bias.data.fill_(0)

    def feature_l1(self, average: bool = False, subs_idx=None) -> torch.Tensor:
        subs_idx = subs_idx if subs_idx is not None else slice(None)
        if average:
            return self.features[..., subs_idx].abs().mean()
        else:
            return self.features[..., subs_idx].abs().sum()

    def mask_l1(self, average: bool = False, subs_idx=None) -> torch.Tensor:
        subs_idx = subs_idx if subs_idx is not None else slice(None)
        if self.gaussian_masks:
            if average:
                return (
                    torch.exp(self.mask_log_var * self.gaussian_var_scale)[..., subs_idx].mean()
                    + (self.mask_mean[..., subs_idx] * self.gaussian_mean_scale).pow(2).mean()
                )
            else:
                return (
                    torch.exp(self.mask_log_var * self.gaussian_var_scale)[..., subs_idx].sum()
                    + (self.mask_mean[..., subs_idx] * self.gaussian_mean_scale).pow(2).sum()
                )
        else:
            if average:
                return self.masks[..., subs_idx].abs().mean()
            else:
                return self.masks[..., subs_idx].abs().sum()

    def make_mask_grid(self, w: int, h: int) -> torch.Tensor:
        """Actually mixed up: w (width) is height, and vice versa"""
        grid_w = torch.linspace(-1 * w / max(w, h), 1 * w / max(w, h), w)
        grid_h = torch.linspace(-1 * h / max(w, h), 1 * h / max(w, h), h)
        xx, yy = torch.meshgrid([grid_w, grid_h], indexing="ij")
        grid = torch.stack([xx, yy], 2)[None, ...]
        return grid.repeat([self.outdims, 1, 1, 1])

    def normal_pdf(self) -> torch.Tensor:
        """Gets the actual mask values in terms of a PDF from the mean and SD"""
        # self.mask_var_ = torch.exp(self.mask_log_var * self.gaussian_var_scale).view(-1, 1, 1)
        scaled_log_var = self.mask_log_var * self.gaussian_var_scale
        self.mask_var_ = torch.exp(torch.clamp(scaled_log_var, min=-20, max=20)).view(-1, 1, 1)
        pdf = self.grid - self.mask_mean.view(self.outdims, 1, 1, -1) * self.gaussian_mean_scale
        pdf = torch.sum(pdf**2, dim=-1) / (self.mask_var_ + 1e-8)
        pdf = torch.exp(-0.5 * torch.clamp(pdf, max=20))
        normalisation = torch.sum(pdf, dim=(1, 2), keepdim=True)
        pdf = torch.nan_to_num(pdf / normalisation)
        return pdf

    def get_masks(self) -> torch.Tensor:
        if self.gaussian_masks:
            return self.normal_pdf().permute(1, 2, 0)
        else:
            return self.masks.data.abs_()

    def forward(self, x: torch.Tensor, shift=None, subs_idx=None) -> torch.Tensor:
        self.masks = self.get_masks()

        if self.positive:
            self.features.data.clamp_(0)

        N, c, t, w, h = x.size()
        if subs_idx is not None:
            feat = self.features[..., subs_idx]
            masks = self.masks[..., subs_idx]

        else:
            feat = self.features
            masks = self.masks

        y = torch.einsum("nctwh,whd->nctd", x, masks)
        y = (y * feat).sum(1)

        if self.scale is not None:
            y = y * self.scale
        if self.bias is not None:
            if subs_idx is None:
                y = y + self.bias
            else:
                y = y + self.bias[subs_idx]
        if self.nonlinearity:
            y = F.softplus(y)
        return y

    def save_weight_visualizations(self, folder_path: str, file_format: str = "jpg", add_titles: bool = True, remove_readout_ticks: bool = False) -> None:
        masks = self.get_masks().detach().cpu().numpy()
        mask_abs_max = np.abs(masks).max()
        features = self.features.detach().cpu().numpy()
        features_min = float(features.min())
        features_max = float(features.max())
        for neuron_id in range(masks.shape[-1]):
            mask_neuron = masks[:, :, neuron_id]
            fig_axes_tuple = plt.subplots(ncols=2, figsize=(2 * 6, 6))
            axes: list[plt.Axes] = fig_axes_tuple[1]  # type: ignore
            axes[0].imshow(
                mask_neuron, interpolation="none", cmap="RdBu_r", norm=Normalize(-mask_abs_max, mask_abs_max)
            )

            features_neuron = features[0, :, 0, neuron_id]
            axes[1].bar(range(features_neuron.shape[0]), features_neuron)
            axes[1].set_ylim((features_min, features_max))

            if add_titles:
                axes[0].set_title("Readout Mask")
                axes[1].set_title("Readout feature weights")
            if remove_readout_ticks:
                axes[1].axes.get_xaxis().set_ticks([])
                axes[1].axes.get_yaxis().set_ticks([])

            plot_path = f"{folder_path}/neuron_{neuron_id}.{file_format}"
            fig_axes_tuple[0].savefig(plot_path, bbox_inches="tight", facecolor="w", dpi=300)
            fig_axes_tuple[0].clf()
            plt.close()

    def __repr__(self) -> str:
        c, _, w, h = self.in_shape
        res_array: list[str] = []
        r = f"{self.__class__.__name__} ( {c} x {w} x {h} -> {str(self.outdims)})"
        if self.bias is not None:
            r += " with bias"
        res_array.append(r)

        for ch in self.children():
            r += "  -> " + ch.__repr__()
            res_array.append(r)
        return "\n".join(res_array)


class SpatialXFeature3dReadout(nn.ModuleDict):
    def __init__(
        self,
        core: Core,
        in_shape_dict,
        n_neurons_dict,
        scale,
        bias,
        gaussian_masks,
        gaussian_mean_scale,
        gaussian_var_scale,
        positive,
        initialize_from_roi_masks,
        roi_masks,
        gamma_readout,
        gamma_masks=0.0,
        readout_reg_avg=False,
    ):
        super().__init__()
        for k in n_neurons_dict:  # iterate over sessions
            n_neurons = n_neurons_dict[k]
            in_shape = get_module_output_shape(core, in_shape_dict[k])[1:]
            assert len(in_shape) == 4
            self.add_module(
                str(k),
                SpatialXFeature3d(  # add a readout for each session
                    in_shape,
                    n_neurons,
                    gaussian_masks=gaussian_masks,
                    gaussian_mean_scale=gaussian_mean_scale,
                    gaussian_var_scale=gaussian_var_scale,
                    initialize_from_roi_masks=initialize_from_roi_masks,
                    roi_mask=roi_masks[k],
                    positive=positive,
                    scale=scale,
                    bias=bias,
                ),
            )

        self.gamma_readout = gamma_readout
        self.gamma_masks = gamma_masks
        self.gaussian_masks = gaussian_masks
        self.readout_reg_avg = readout_reg_avg

    def forward(self, *args, data_key=None, **kwargs):
        if data_key is None and len(self) == 1:
            data_key = list(self.keys())[0]
        return self[data_key](*args, **kwargs)

    def regularizer(self, data_key):
        ret = self[data_key].feature_l1(average=self.readout_reg_avg) * self.gamma_readout
        ret = ret + self[data_key].mask_l1(average=self.readout_reg_avg) * self.gamma_masks
        return ret

    def readout_keys(self) -> list[str]:
        return sorted(self._modules.keys())


class LocalEncoder(Encoder):
    def forward(
        self, x: torch.Tensor, data_key: str | None = None, detach_core: bool = False, **kwargs
    ) -> torch.Tensor:
        self.detach_core = detach_core
        if self.detach_core:
            for name, param in self.core.features.named_parameters():
                if name.find("speed") < 0:
                    param.requires_grad = False
        x = self.core(x, data_key=data_key)
        x = self.readout(x, data_key=data_key)
        return x

    @staticmethod
    def stimulus_shape(time_steps: int, num_batches: int = 1) -> tuple[int, int, int, int, int]:
        channels, width, height = 2, 18, 16
        return num_batches, channels, time_steps, width, height


# Batch adaption model
def SFB3d_core_SxF3d_readout(
    dataloaders,
    seed: int | None = None,
    hidden_channels: Tuple[int] = (8,),  # core args
    temporal_kernel_size: Tuple[int] = (21,),
    spatial_kernel_size: Tuple[int] = (11,),
    layers: int = 1,
    gamma_hidden: float = 0,
    gamma_input: float = 0.1,
    gamma_temporal: float = 0.1,
    gamma_in_sparse: float = 0.0,
    final_nonlinearity: bool = True,
    core_bias: bool = False,
    momentum: float = 0.1,
    input_padding: bool = False,
    hidden_padding: bool = True,
    batch_norm: bool = True,
    batch_norm_scale: bool = False,
    laplace_padding=None,
    batch_adaptation: bool = True,
    readout_scale: bool = False,
    readout_bias: bool = True,
    gaussian_masks: bool = False,  # readout args,
    gamma_readout: float = 0.1,
    gamma_masks: float = 0,
    gaussian_mean_scale: float = 1e0,
    gaussian_var_scale: float = 1e0,
    initialize_from_roi_masks: bool = False,
    readout_positive: bool = False,
    stack=None,
    readout_reg_avg: bool = False,
    use_avg_reg: bool = False,
    data_info: Optional[dict] = None,
    nonlinearity: str = "ELU",
    conv_type: Literal["full", "separable", "custom_separable", "time_independent"] = "custom_separable",
) -> torch.nn.Module:
    """
    Model class of a stacked2dCore (from mlutils) and a pointpooled (spatial transformer) readout
    Args:
        dataloaders: a dictionary of dataloaders, one loader per sessionin the format:
            {'train': {'session1': dataloader1, 'session2': dataloader2, ...},
             'validation': {'session1': dataloader1, 'session2': dataloader2, ...},
             'test': {'session1': dataloader1, 'session2': dataloader2, ...}}
        seed: random seed
        elu_offset: Offset for the output non-linearity [F.elu(x + self.offset)]
        all other args: See Documentation of Stacked2dCore in mlutils.layers.cores and
            PointPooled2D in mlutils.layers.readouts
    Returns: An initialized model which consists of model.core and model.readout
    """

    # make sure trainloader is being used
    if data_info is not None:
        in_shapes_dict = {k: v["input_dimensions"] for k, v in data_info.items()}
        input_channels = [v["input_channels"] for k, v in data_info.items()]
        n_neurons_dict = {k: v["output_dimension"] for k, v in data_info.items()}
        roi_masks = {k: torch.tensor(v["roi_coords"]) for k, v in data_info.items()}
    else:
        dataloaders = dataloaders.get("train", dataloaders)

        # Obtain the named tuple fields from the first entry of the first dataloader in the dictionary
        in_name, out_name, *_ = next(iter(list(dataloaders.values())[0]))._fields

        session_shape_dict = get_dims_for_loader_dict(dataloaders)
        print(session_shape_dict)
        n_neurons_dict = {
            k: v[out_name][-1] for k, v in session_shape_dict.items()
        }  # dictionary containing # neurons per session
        in_shapes_dict = {
            k: v[in_name] for k, v in session_shape_dict.items()
        }  # dictionary containing input shapes per session
        input_channels = [v[in_name][1] for v in session_shape_dict.values()]  # gets the # of input channels
        roi_masks = {k: dataloaders[k].dataset.roi_coords for k in dataloaders.keys()}
    assert np.unique(input_channels).size == 1, "all input channels must be of equal size"

    set_seed(seed)

    # get a stacked factorized 3d core from below
    core = ParametricFactorizedBatchConv3dCore(
        n_neurons_dict=n_neurons_dict,
        input_channels=input_channels[0],
        num_scans=len(n_neurons_dict.keys()),
        hidden_channels=hidden_channels,
        temporal_kernel_size=temporal_kernel_size,
        spatial_kernel_size=spatial_kernel_size,
        layers=layers,
        gamma_hidden=gamma_hidden,
        gamma_input=gamma_input,
        gamma_in_sparse=gamma_in_sparse,
        gamma_temporal=gamma_temporal,
        final_nonlinearity=final_nonlinearity,
        bias=core_bias,
        momentum=momentum,
        input_padding=input_padding,
        hidden_padding=hidden_padding,
        batch_norm=batch_norm,
        batch_norm_scale=batch_norm_scale,
        laplace_padding=laplace_padding,
        stack=stack,
        batch_adaptation=batch_adaptation,
        use_avg_reg=use_avg_reg,
        nonlinearity=nonlinearity,
        conv_type=conv_type,
    )

    readout = SpatialXFeature3dReadout(
        core,
        in_shape_dict=in_shapes_dict,
        n_neurons_dict=n_neurons_dict,
        scale=readout_scale,
        bias=readout_bias,
        gaussian_masks=gaussian_masks,
        gaussian_mean_scale=gaussian_mean_scale,
        gaussian_var_scale=gaussian_var_scale,
        positive=readout_positive,
        initialize_from_roi_masks=initialize_from_roi_masks,
        roi_masks=roi_masks,
        gamma_readout=gamma_readout,
        gamma_masks=gamma_masks,
        readout_reg_avg=readout_reg_avg,
    )

    # initializing readout bias to mean response
    if readout_bias:
        if data_info is None:
            for k in dataloaders:
                readout[k].bias.data = dataloaders[k].dataset.mean_response
        else:
            for k in data_info.keys():
                readout[k].bias.data = torch.from_numpy(data_info[k]["mean_response"])

    model = LocalEncoder(
        core,
        readout,
    )

    return model



# ruff: noqa: E501  # for long link in this file, it didn't work to put it at that specific line for some reason
from collections import OrderedDict
from collections.abc import Iterable
from operator import itemgetter
from typing import Any, Dict, Literal, Optional, Tuple

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from neuralpredictors.layers.rnn_modules.gru_module import ConvGRUCell  # type: ignore
from neuralpredictors.utils import get_module_output  # type: ignore

from openretina.data_io.hoefling_2024.dataloaders import get_dims_for_loader_dict
from openretina.legacy.models import (
    Bias3DLayer,
    FlatLaplaceL23dnorm,
    Scale2DLayer,
    Scale3DLayer,
    TimeLaplaceL23dnorm,
)
from openretina.models.dev import Encoder
from openretina.models.dev import MultiGaussian2d
from openretina.modules.core.base_core import Core3d
from openretina.modules.layers.convolutions import STSeparableBatchConv3d, TimeIndependentConv3D, TorchFullConv3D, TorchSTSeparableConv3D, compute_temporal_kernel, temporal_smoothing
from openretina.utils.misc import set_seed

DEVICE = "cuda:0" if torch.cuda.is_available() else "cpu"


class GRUEnabledCore(Core3d, nn.Module):
    def __init__(
        self,
        n_neurons_dict,
        input_channels=2,
        num_scans=1,
        hidden_channels=[8],
        temporal_kernel_size=[21],
        spatial_kernel_size=[14],
        layers=1,
        gamma_hidden=0.0,
        gamma_input=0.0,
        gamma_in_sparse=0.0,
        gamma_temporal=0.0,
        final_nonlinearity=True,
        bias=True,
        momentum=0.1,
        input_padding=False,
        hidden_padding=True,
        batch_norm=True,
        batch_norm_scale=True,
        laplace_padding=0,
        stack=None,
        batch_adaptation=True,
        use_avg_reg=False,
        nonlinearity="ELU",
        conv_type="custom_separable",
        use_gru=False,
        device=DEVICE,
        gru_kwargs: Optional[Dict[str, int | float]] = None,
    ):
        super().__init__()
        self._input_weights_regularizer_spatial = FlatLaplaceL23dnorm(padding=laplace_padding)
        self._input_weights_regularizer_temporal = TimeLaplaceL23dnorm(padding=laplace_padding)

        if conv_type == "separable":
            self.conv_class = TorchSTSeparableConv3D  # type: ignore
        elif conv_type == "custom_separable":
            self.conv_class = STSeparableBatchConv3d  # type: ignore
        elif conv_type == "full":
            self.conv_class = TorchFullConv3D  # type: ignore
        elif conv_type == "time_independent":
            self.conv_class = TimeIndependentConv3D  # type: ignore
        else:
            raise ValueError(f"Un-implemented conv_type {conv_type}")

        self.num_scans = num_scans
        self.layers = layers
        self.gamma_input = gamma_input
        self.gamma_in_sparse = gamma_in_sparse
        self.gamma_hidden = gamma_hidden
        self.gamma_temporal = gamma_temporal
        self.input_channels = input_channels
        self.hidden_channels = hidden_channels
        self.use_avg_reg = use_avg_reg

        self.features = nn.Sequential()
        if stack is None:
            self.stack = list(range(self.layers))
        else:
            self.stack = [range(self.layers)[stack]] if isinstance(stack, int) else stack  # type: ignore

        log_speed_dict = dict()
        for k in n_neurons_dict:
            var_name = "_".join(["log_speed", k])
            log_speed_val = torch.nn.Parameter(data=torch.zeros(1), requires_grad=batch_adaptation)
            setattr(self, var_name, log_speed_val)
            log_speed_dict[var_name] = log_speed_val

        if input_padding:
            input_pad: Tuple[int, ...] | int = (0, spatial_kernel_size[0] // 2, spatial_kernel_size[0] // 2)
        else:
            input_pad = 0
        if hidden_padding & (len(spatial_kernel_size) > 1):
            hidden_pad: list[Tuple[int, ...] | int] = [
                (0, spatial_kernel_size[x] // 2, spatial_kernel_size[x] // 2)
                for x in range(1, len(spatial_kernel_size))
            ]
        else:
            hidden_pad = [0 for _ in range(1, len(spatial_kernel_size))]

        if not isinstance(hidden_channels, Iterable):
            hidden_channels = [hidden_channels] * (self.layers)
        if not isinstance(temporal_kernel_size, Iterable):
            temporal_kernel_size = [temporal_kernel_size] * (self.layers)
        if not isinstance(spatial_kernel_size, Iterable):
            spatial_kernel_size = [spatial_kernel_size] * (self.layers)

        # --- first layer
        layer: OrderedDict[str, Any] = OrderedDict()
        layer["conv"] = self.conv_class(
            input_channels,
            hidden_channels[0],
            log_speed_dict,
            temporal_kernel_size[0],
            spatial_kernel_size[0],
            bias=False,
            padding=input_pad,  # type: ignore
            num_scans=self.num_scans,
        )
        if batch_norm:
            layer["norm"] = nn.BatchNorm3d(
                hidden_channels[0],  # type: ignore
                momentum=momentum,
                affine=bias and batch_norm_scale,
            )  # ok or should we ensure same batch norm?
            if bias:
                if not batch_norm_scale:
                    layer["bias"] = Bias3DLayer(hidden_channels[0])
            elif batch_norm_scale:
                layer["scale"] = Scale3DLayer(hidden_channels[0])
        if final_nonlinearity:
            layer["nonlin"] = getattr(nn, nonlinearity)()  # TODO add back in place if necessary
        self.features.add_module("layer0", nn.Sequential(layer))  # type: ignore

        # --- other layers

        for layer_num in range(1, self.layers):
            layer = OrderedDict()
            layer["conv"] = self.conv_class(
                hidden_channels[layer_num - 1],
                hidden_channels[layer_num],
                log_speed_dict,
                temporal_kernel_size[layer_num],
                spatial_kernel_size[layer_num],
                bias=False,
                padding=hidden_pad[layer_num - 1],  # type: ignore
                num_scans=self.num_scans,
            )
            if batch_norm:
                layer["norm"] = nn.BatchNorm3d(
                    hidden_channels[layer_num],
                    momentum=momentum,
                    affine=bias and batch_norm_scale,
                )
                if bias:
                    if not batch_norm_scale:
                        layer["bias"] = Bias3DLayer(hidden_channels[layer_num])
                elif batch_norm_scale:
                    layer["scale"] = Scale2DLayer(hidden_channels[layer_num])
            if final_nonlinearity or layer_num < self.layers - 1:
                layer["nonlin"] = getattr(nn, nonlinearity)()
            self.features.add_module(f"layer{layer_num}", nn.Sequential(layer))

        self.apply(self.init_conv)

        if use_gru:
            print("Using GRU")
            self.features.add_module("gru", GRU_Module(**gru_kwargs))  # type: ignore

    def forward(self, input_, data_key=None):
        ret = []
        for layer_num, feat in enumerate(self.features):
            do_skip = False
            input_ = feat(
                (
                    input_ if not do_skip else torch.cat(ret[-min(self.skip, layer_num) :], dim=1),
                    data_key,
                )
            )
            ret.append(input_)

        return torch.cat([ret[ind] for ind in self.stack], dim=1)

    def spatial_laplace(self):
        return self._input_weights_regularizer_spatial(self.features[0].conv.weight_spatial, avg=self.use_avg_reg)

    def group_sparsity(self):  # check if this is really what we want
        sparsity_loss = 0
        for layer_num in range(1, self.layers):
            spatial_weight_layer = self.features[layer_num].conv.weight_spatial
            norm = spatial_weight_layer.pow(2).sum([2, 3, 4]).sqrt().sum(1)
            sparsity_loss_layer = (spatial_weight_layer.pow(2).sum([2, 3, 4]).sqrt().sum(1) / norm).sum()
            sparsity_loss += sparsity_loss_layer
        return sparsity_loss

    def group_sparsity0(self):  # check if this is really what we want
        weight_temporal = compute_temporal_kernel(
            torch.zeros(1, device=self.features[0].conv.sin_weights.device),
            self.features[0].conv.sin_weights,
            self.features[0].conv.cos_weights,
            self.features[0].conv.temporal_kernel_size,
        )
        # abc are dummy dimensions
        weight = torch.einsum("oitab,oichw->oithw", weight_temporal, self.features[0].conv.weight_spatial)
        ret = (weight.pow(2).sum([2, 3, 4]).sqrt().sum(1) / torch.sqrt(1e-8 + weight.pow(2).sum([1, 2, 3, 4]))).sum()
        return ret

    def temporal_smoothness(self):
        ret = 0
        for layer_num in range(self.layers):
            ret += temporal_smoothing(
                self.features[layer_num].conv.sin_weights, self.features[layer_num].conv.cos_weights
            )
        return ret

    def regularizer(self):
        if self.conv_class == STSeparableBatchConv3d:
            return (
                self.group_sparsity() * self.gamma_hidden
                + self.gamma_input * self.spatial_laplace()
                + self.gamma_temporal * self.temporal_smoothness()
                + self.group_sparsity0() * self.gamma_in_sparse
            )
        else:
            return 0

    @property
    def outchannels(self):
        return len(self.features) * self.hidden_channels[-1]


class VideoEncoder(Encoder):
    def forward(self, x, data_key=None, detach_core=False, **kwargs):
        self.detach_core = detach_core
        if self.detach_core:
            for name, param in self.core.features.named_parameters():
                if name.find("speed") < 0:
                    param.requires_grad = False

        x = self.core(x, data_key=data_key)

        # Make time the second dimension again for the readout
        x = torch.transpose(x, 1, 2)

        # Get dims for later reshaping
        batch_size = x.shape[0]
        time_points = x.shape[1]

        # Treat time as an indipendent (batch) dimension for the readout
        x = x.reshape(((-1,) + x.size()[2:]))
        x = self.readout(x, data_key=data_key, **kwargs)

        # Reshape back to the correct dimensions before returning
        x = x.reshape(((batch_size, time_points) + x.size()[1:]))

        # Add softplus as gaussian readout does not assure positive values
        x = F.softplus(x)
        return x


class GRU_Module(nn.Module):
    def __init__(
        self,
        input_channels,
        rec_channels,
        input_kern,
        rec_kern,
        groups=1,
        gamma_rec=0,
        pad_input=True,
        **kwargs,
    ):
        """
        A GRU module for video data to add between the core and the readout.
        Receives as input the output of a 3Dcore. Expected dimensions:
            - (Batch, Channels, Frames, Height, Width) or (Channels, Frames, Height, Width)
        The input is fed sequentially to a convolutional GRU cell, based on the frames channel.
        The output has the same dimensions as the input.
        """
        super().__init__()
        self.gru = ConvGRUCell(
            input_channels,
            rec_channels,
            input_kern,
            rec_kern,
            groups=groups,
            gamma_rec=gamma_rec,
            pad_input=pad_input,
        )

    def forward(self, input_):
        """
        Forward pass definition based on
        https://github.com/sinzlab/Sinz2018_NIPS/blob/3a99f7a6985ae8dec17a5f2c54f550c2cbf74263/nips2018/architectures/cores.py#L556
        Modified to also accept 4 dimensional inputs (assuming no batch dimension is provided).
        """
        x, data_key = input_
        if len(x.shape) not in [4, 5]:
            raise RuntimeError(
                f"Expected 4D (unbatched) or 5D (batched) input to ConvGRUCell, but got input of size: {x.shape}"
            )

        batch = True
        if len(x.shape) == 4:
            batch = False
            x = torch.unsqueeze(x, dim=0)

        states = []
        hidden = None
        frame_pos = 2

        for frame in range(x.shape[frame_pos]):
            slice_channel = [frame if frame_pos == i else slice(None) for i in range(len(x.shape))]
            hidden = self.gru(x[slice_channel], hidden)
            states.append(hidden)
        out = torch.stack(states, frame_pos)
        if not batch:
            out = torch.squeeze(out, dim=0)
        return out


# Gaussian readout model
def SFB3d_core_gaussian_readout(
    dataloaders,
    seed,
    hidden_channels: Tuple[int] = (8,),  # core args
    temporal_kernel_size: Tuple[int] = (21,),
    spatial_kernel_size: Tuple[int] = (11,),
    layers: int = 1,
    gamma_hidden: float = 0,
    gamma_input: float = 0.1,
    gamma_temporal: float = 0.1,
    gamma_in_sparse=0.0,
    final_nonlinearity: bool = True,
    core_bias: bool = False,
    momentum: float = 0.1,
    input_padding: bool = False,
    hidden_padding: bool = True,
    batch_norm: bool = True,
    batch_norm_scale: bool = False,
    laplace_padding=None,
    batch_adaptation: bool = True,
    readout_scale: bool = False,
    readout_bias: bool = True,
    gaussian_masks: bool = False,  # readout args,
    gamma_readout: float = 0.1,
    gamma_masks: float = 0,
    gaussian_mean_scale: float = 1e0,
    gaussian_var_scale: float = 1e0,
    initialize_from_roi_masks: bool = False,
    readout_positive: bool = False,
    stack=None,
    readout_reg_avg: bool = False,
    use_avg_reg: bool = False,
    data_info: Optional[dict] = None,
    nonlinearity: str = "ELU",
    conv_type: Literal["full", "separable", "custom_separable", "time_independent"] = "custom_separable",
    device=DEVICE,
    use_gru: bool = False,
    gru_kwargs: dict = {},
):
    """
    Model class of a stacked2dCore (from mlutils) and a pointpooled (spatial transformer) readout
    Args:
        dataloaders: a dictionary of dataloaders, one loader per sessionin the format:
            {'train': {'session1': dataloader1, 'session2': dataloader2, ...},
             'validation': {'session1': dataloader1, 'session2': dataloader2, ...},
             'test': {'session1': dataloader1, 'session2': dataloader2, ...}}
        seed: random seed
        elu_offset: Offset for the output non-linearity [F.elu(x + self.offset)]
        all other args: See Documentation of Stacked2dCore in mlutils.layers.cores and
            PointPooled2D in mlutils.layers.readouts
    Returns: An initialized model which consists of model.core and model.readout
    """

    # make sure trainloader is being used
    if data_info is not None:
        in_shapes_dict = {k: v["input_dimensions"] for k, v in data_info.items()}
        input_channels = [v["input_channels"] for k, v in data_info.items()]
        n_neurons_dict = {k: v["output_dimension"] for k, v in data_info.items()}
    else:
        dataloaders = dataloaders.get("train", dataloaders)

        # Obtain the named tuple fields from the first entry of the first dataloader in the dictionary
        in_name, out_name, *_ = next(iter(list(dataloaders.values())[0]))._fields

        session_shape_dict = get_dims_for_loader_dict(dataloaders)
        print(session_shape_dict)
        n_neurons_dict = {
            k: v[out_name][-1] for k, v in session_shape_dict.items()
        }  # dictionary containing # neurons per session
        in_shapes_dict = {
            k: v[in_name] for k, v in session_shape_dict.items()
        }  # dictionary containing input shapes per session
        input_channels = [v[in_name][1] for v in session_shape_dict.values()]  # gets the # of input channels
    assert np.unique(input_channels).size == 1, "all input channels must be of equal size"

    set_seed(seed)

    # get a stacked factorized 3d core from below
    core = GRUEnabledCore(
        n_neurons_dict=n_neurons_dict,
        input_channels=input_channels[0],
        num_scans=len(n_neurons_dict.keys()),
        hidden_channels=hidden_channels,
        temporal_kernel_size=temporal_kernel_size,
        spatial_kernel_size=spatial_kernel_size,
        layers=layers,
        gamma_hidden=gamma_hidden,
        gamma_input=gamma_input,
        gamma_in_sparse=gamma_in_sparse,
        gamma_temporal=gamma_temporal,
        final_nonlinearity=final_nonlinearity,
        bias=core_bias,
        momentum=momentum,
        input_padding=input_padding,
        hidden_padding=hidden_padding,
        batch_norm=batch_norm,
        batch_norm_scale=batch_norm_scale,
        laplace_padding=laplace_padding,
        stack=stack,
        batch_adaptation=batch_adaptation,
        use_avg_reg=use_avg_reg,
        nonlinearity=nonlinearity,
        conv_type=conv_type,
        device=device,
        use_gru=use_gru,
        gru_kwargs=gru_kwargs,
    )

    in_shapes_readout = {}
    subselect = itemgetter(0, 2, 3)
    for k in n_neurons_dict:  # iterate over sessions
        in_shapes_readout[k] = subselect(tuple(get_module_output(core, in_shapes_dict[k])[1:]))

    readout = MultiGaussian2d(
        in_shape_dict=in_shapes_readout,
        n_neurons_dict=n_neurons_dict,
        scale=readout_scale,
        bias=readout_bias,
        feature_reg_weights=gamma_readout,
    )

    # initializing readout bias to mean response
    if readout_bias is True:
        if data_info is None:
            for k in dataloaders:
                readout[k].bias.data = dataloaders[k].dataset[:]._asdict()[out_name].mean(0)
        else:
            for k in data_info.keys():
                readout[k].bias.data = torch.from_numpy(data_info[k]["mean_response"])

    model = VideoEncoder(
        core,
        readout,
    )

    return model






import torch
from torch import nn


class CorrelationLoss3d(nn.Module):
    def __init__(self, bias: float = 1e-16, per_neuron: bool = False, avg: bool = False):
        super().__init__()
        self.eps = bias
        self.per_neuron = per_neuron
        self.avg = avg

    def forward(self, output, target):
        lag = target.size(1) - output.size(1)
        delta_out = output - output.mean(1, keepdim=True)
        delta_target = target[:, lag:, :] - target[:, lag:, :].mean(1, keepdim=True)
        var_out = delta_out.pow(2).mean(1, keepdim=True)
        var_target = delta_target.pow(2).mean(1, keepdim=True)
        corrs = (delta_out * delta_target).mean(1, keepdim=True) / (
            (var_out + self.eps) * (var_target + self.eps)
        ).sqrt()
        if not self.per_neuron:
            return -corrs.mean() if self.avg else -corrs.sum()
        else:
            return -corrs.view(-1, corrs.shape[-1]).mean(dim=0)


class CelltypeCorrelationLoss3d(nn.Module):
    def __init__(self, bias: float = 1e-16, per_neuron: bool = False, avg: bool = False):
        super().__init__()
        self.eps = bias
        self.per_neuron = per_neuron
        self.avg = avg

    def forward(self, output, target, group_assignment, group_counts):
        lag = target.size(1) - output.size(1)
        delta_out = output - output.mean(1, keepdim=True)
        delta_target = target[:, lag:, :] - target[:, lag:, :].mean(1, keepdim=True)
        var_out = delta_out.pow(2).mean(1, keepdim=True)
        var_target = delta_target.pow(2).mean(1, keepdim=True)
        corrs = (delta_out * delta_target).mean(1, keepdim=True) / (
            (var_out + self.eps) * (var_target + self.eps)
        ).sqrt()
        corrs = corrs * (group_counts.sum() / group_counts.size(0) / group_counts[group_assignment - 1])[:, None, ...]
        if not self.per_neuron:
            return -corrs.mean() if self.avg else -corrs.sum()
        else:
            return -corrs.view(-1, corrs.shape[-1]).mean(dim=0)


class L1CorrelationLoss3d(nn.Module):
    def __init__(self, bias: float = 1e-16, per_neuron: bool = False, avg: bool = False):
        super().__init__()
        self.eps = bias
        self.per_neuron = per_neuron
        self.avg = avg
        self.gamma_L1 = 0.0002

    def forward(self, output, target, **kwargs):
        pre_ca = output[1]
        output = output[0]
        lag = target.size(1) - output.size(1)
        delta_out = output - output.mean(1, keepdim=True)
        delta_target = target[:, lag:, :] - target[:, lag:, :].mean(1, keepdim=True)
        var_out = delta_out.pow(2).mean(1, keepdim=True)
        var_target = delta_target.pow(2).mean(1, keepdim=True)
        corrs = (delta_out * delta_target).mean(1, keepdim=True) / (
            (var_out + self.eps) * (var_target + self.eps)
        ).sqrt()
        if not self.per_neuron:
            ret1 = -corrs.mean() if self.avg else -corrs.sum()
            ret2 = self.gamma_L1 * (pre_ca.abs().mean() if self.avg else pre_ca.abs().sum())
        else:
            ret1 = -corrs.view(-1, corrs.shape[-1]).mean(dim=0)
            pre_ca = pre_ca.transpose(1, 2)
            ret2 = self.gamma_L1 * (pre_ca.view(-1, pre_ca.shape[-1]).abs().mean(dim=0))
        print("corr loss", ret1)
        print("L1 loss", ret2)
        return ret1 + ret2


class ScaledCorrelationLoss3d(nn.Module):
    def __init__(self, bias=1e-16, scale=30, per_neuron=False, avg=False):
        super().__init__()
        self.eps = bias
        self.scale = scale
        self.per_neuron = per_neuron
        self.avg = avg

    def forward(self, output, target, **kwargs):
        lag = target.size(1) - output.size(1)
        corrs = torch.zeros((output.size(0), 1, output.size(2)), device=output.device)
        count = 0
        for i in range(output.size(1) // self.scale):
            delta_out = output[:, i * self.scale : (i + 1) * self.scale, :] - output[
                :, i * self.scale : (i + 1) * self.scale, :
            ].mean(1, keepdim=True)
            delta_target = target[:, lag + i * self.scale : lag + (i + 1) * self.scale, :] - target[
                :, lag + i * self.scale : lag + (i + 1) * self.scale, :
            ].mean(1, keepdim=True)
            var_out = delta_out.pow(2).mean(1, keepdim=True)
            var_target = delta_target.pow(2).mean(1, keepdim=True)
            corrs += (delta_out * delta_target).mean(1, keepdim=True) / (
                (var_out + self.eps) * (var_target + self.eps)
            ).sqrt()
            count += 1
        corrs /= count
        if not self.per_neuron:
            return -corrs.mean() if self.avg else -corrs.sum()
        else:
            return -corrs.view(-1, corrs.shape[-1]).mean(dim=0)



import matplotlib.pyplot as plt
import numpy as np
import torch
from matplotlib.colors import Normalize
from torch import nn


class SimpleSpatialXFeature3d(torch.nn.Module):
    def __init__(
        self,
        in_shape: tuple[int, int, int, int],
        outdims: int,
        gaussian_mean_scale: float = 1e0,
        gaussian_var_scale: float = 1e0,
        positive: bool = False,
        scale: bool = False,
        bias: bool = True,
        nonlinearity_function=torch.nn.functional.softplus,
    ):
        """
        Args:
            in_shape: The shape of the input tensor (c, t, w, h).
            outdims: The number of output dimensions (usually the number of neurons in the session).
            gaussian_mean_scale: The scale factor for the Gaussian mean. Defaults to 1e0.
            gaussian_var_scale: The scale factor for the Gaussian variance. Defaults to 1e0.
            positive: Whether the output should be positive. Defaults to False.
            scale: Whether to include a scale parameter. Defaults to False.
            bias: Whether to include a bias parameter. Defaults to True.
            nonlinearity_function: torch nonlinearity function , e.g. nn.functional.softplus
        """
        super().__init__()
        self.in_shape = in_shape
        c, t, w, h = in_shape
        self.outdims = outdims
        self.gaussian_mean_scale = gaussian_mean_scale
        self.gaussian_var_scale = gaussian_var_scale
        self.positive = positive
        self.nonlinearity_function = nonlinearity_function

        """we train on the log var and transform to var in a separate step"""
        self.mask_mean = torch.nn.Parameter(data=torch.zeros(self.outdims, 2), requires_grad=True)
        self.mask_log_var = torch.nn.Parameter(data=torch.zeros(self.outdims), requires_grad=True)

        # Grid is fixed and untrainable, so we register it as a buffer
        self.register_buffer("grid", self.make_mask_grid(outdims, w, h))

        self.features = nn.Parameter(torch.Tensor(1, c, 1, outdims))
        self.features.data.normal_(1.0 / c, 0.01)

        if scale:
            self.scale_param = nn.Parameter(torch.ones(outdims))
            self.scale_param.data.normal_(1.0, 0.01)
        else:
            self.register_buffer("scale_param", torch.ones(outdims))  # Non-trainable

        if bias:
            self.bias_param = nn.Parameter(torch.zeros(outdims))
        else:
            self.register_buffer("bias_param", torch.zeros(outdims))  # Non-trainable

    def feature_l1(self, average: bool = False) -> torch.Tensor:
        features_abs = self.features.abs()
        if average:
            return features_abs.mean()
        else:
            return features_abs.sum()

    def mask_l1(self, average: bool = False) -> torch.Tensor:
        if average:
            return (
                torch.exp(self.mask_log_var * self.gaussian_var_scale).mean()
                + (self.mask_mean * self.gaussian_mean_scale).pow(2).mean()
            )
        else:
            return (
                torch.exp(self.mask_log_var * self.gaussian_var_scale).sum()
                + (self.mask_mean * self.gaussian_mean_scale).pow(2).sum()
            )

    @staticmethod
    def make_mask_grid(outdims: int, w: int, h: int) -> torch.Tensor:
        """Actually mixed up: w (width) is height, and vice versa"""
        grid_w = torch.linspace(-1 * w / max(w, h), 1 * w / max(w, h), w)
        grid_h = torch.linspace(-1 * h / max(w, h), 1 * h / max(w, h), h)
        xx, yy = torch.meshgrid([grid_w, grid_h], indexing="ij")
        grid = torch.stack([xx, yy], 2)[None, ...]
        return grid.repeat([outdims, 1, 1, 1])

    def get_mask(self) -> torch.Tensor:
        """Gets the actual mask values in terms of a PDF from the mean and SD"""
        scaled_log_var = self.mask_log_var * self.gaussian_var_scale
        mask_var_ = torch.exp(torch.clamp(scaled_log_var, min=-20, max=20)).view(-1, 1, 1)
        pdf = self.grid - self.mask_mean.view(self.outdims, 1, 1, -1) * self.gaussian_mean_scale
        pdf = torch.sum(pdf**2, dim=-1) / (mask_var_ + 1e-8)
        pdf = torch.exp(-0.5 * torch.clamp(pdf, max=20))
        normalisation = torch.sum(pdf, dim=(1, 2), keepdim=True)
        pdf = torch.nan_to_num(pdf / normalisation)
        return pdf

    @property
    def masks(self) -> torch.Tensor:
        return self.get_mask().permute(1, 2, 0)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        masks = self.masks
        y = torch.einsum("nctwh,whd->nctd", x, masks)

        if self.positive:
            self.features.data.clamp_(0)
        y = (y * self.features).sum(1)

        y = self.nonlinearity_function(y * self.scale_param + self.bias_param)
        return y

    def __repr__(self) -> str:
        c, _, w, h = self.in_shape
        res_array: list[str] = []
        r = f"{self.__class__.__name__} ( {c} x {w} x {h} -> {str(self.outdims)})"
        if not self.bias_param.requires_grad:
            r += " with bias"
        res_array.append(r)

        children_string = "".join([f" -> {ch.__repr__}" for ch in self.children()])
        res_array.append(children_string)
        return "\n".join(res_array)

    def plot_weight_for_neuron(
        self,
        neuron_id: int,
        axes: tuple[plt.Axes, plt.Axes] | None = None,
        add_titles: bool = True,
        remove_readout_ticks: bool = False,
    ) -> plt.Figure:
        if axes is None:
            fig_axes_tuple = plt.subplots(ncols=2, figsize=(2 * 6, 6))
            ax_readout, ax_features = fig_axes_tuple[1]  # type: ignore
        else:
            ax_readout, ax_features = axes

        masks = self.get_mask().detach().cpu().numpy()
        features = self.features.detach().cpu().numpy()
        mask_abs_max = np.absolute(masks).max()
        mask_neuron = masks[neuron_id, :, :]

        ax_readout.imshow(mask_neuron, interpolation="none", cmap="RdBu_r", norm=Normalize(-mask_abs_max, mask_abs_max))
        ax_readout.axes.get_xaxis().set_ticks([])
        ax_readout.axes.get_yaxis().set_ticks([])

        features_neuron = features[0, :, 0, neuron_id]
        ax_features.bar(range(features_neuron.shape[0]), features_neuron)
        ax_features.set_ylim((features.min(), features.max()))
        if remove_readout_ticks:
            ax_features.axes.get_xaxis().set_ticks([])
            ax_features.axes.get_yaxis().set_ticks([])

        if add_titles:
            ax_readout.set_title("Readout Mask")
            ax_features.set_title("Readout Feature Weights")

        return plt.gcf()

    def save_weight_visualizations(self, folder_path: str, file_format: str = "jpg") -> None:
        for neuron_id in range(self.outdims):
            fig_axes_tuple = plt.subplots(ncols=2, figsize=(2 * 6, 6))
            axes: tuple[plt.Axes, plt.Axes] = fig_axes_tuple[1]  # type: ignore
            self.plot_weight_for_neuron(neuron_id, axes)

            plot_path = f"{folder_path}/neuron_{neuron_id}.{file_format}"
            fig_axes_tuple[0].savefig(plot_path, bbox_inches="tight", facecolor="w", dpi=300)
            fig_axes_tuple[0].clf()
            plt.close()



import logging
import os.path
from pathlib import Path
from typing import Iterable

import h5py
import numpy as np
import torch
from jaxtyping import Float

from openretina.data_io.base import MoviesTrainTestSplit, ResponsesTrainTestSplit

LOGGER = logging.getLogger(__name__)

_STIMULUS_FOLDER = "stimuli"
_RESPONSES_PREFIX = "responses_"
_SESSION_SPECIFIC_STIMULUS_PREFIX = "stimulus_"
_SESSION_INFO_KEY = "session_info"


class TrainTestStimuliProcessor:
    """Manages the processing of stimuli. Useful for caching results"""

    def __init__(
        self,
        test_stimuli: Iterable[str],
        name_to_stimulus: dict[str, np.ndarray],
        norm_mean: float | None,
        norm_std: float | None,
        train_stimuli: Iterable[str] | None = None,
    ):
        self._test_stimuli = set(test_stimuli)
        self._name_to_stimulus = name_to_stimulus
        self._norm_mean = 0.0 if norm_mean is None else norm_mean
        self._norm_std = 1.0 if norm_std is None else norm_std
        self._train_stimuli = set(train_stimuli) if train_stimuli is not None else None

    def _normalize_stimulus(self, stimulus: np.ndarray) -> np.ndarray:
        # Using pytorch is faster than numpy
        stimulus_tensor = torch.tensor(stimulus, dtype=torch.float32)
        normalized_tensor = (stimulus_tensor - self._norm_mean) / self._norm_std
        return normalized_tensor.detach().numpy()

    def process(
        self, stimulus_names: Iterable[str], session_specific_stimuli: dict[str, np.ndarray]
    ) -> MoviesTrainTestSplit:
        # Todo: maybe add caching in the future to reduce memory consumption
        train_stimulus_array, test_stimulus_dict = [], {}

        for name in stimulus_names:
            stimulus = session_specific_stimuli.get(name, self._name_to_stimulus[name])
            normalized_stimulus = self._normalize_stimulus(stimulus)
            if name in self._test_stimuli:
                test_stimulus_dict[name] = normalized_stimulus
            else:
                train_stimulus_array.append(normalized_stimulus)
        # concatenate stimuli over time dimension
        train_stimuli = np.concatenate(train_stimulus_array, axis=1)

        return MoviesTrainTestSplit(train_stimuli, test_stimulus_dict)


def _check_stimulus_size(
    stimulus: Float[np.ndarray, "channels time height width"],
    stimulus_size: tuple[int, int, int],
    stimulus_name: str,
) -> bool:
    """Expecting Float[np.ndarray, "channels time height width"] with height, width == stimulus_size[-2:]
    and channels == stimulus_size[0]
    """
    if len(stimulus.shape) != 4:
        raise ValueError(
            f"Expected four dimensional stimulus (channels, time, height, width), "
            f" but stimulus {stimulus_name} has {stimulus.shape=}"
        )
    elif (stimulus.shape[0] != stimulus_size[0]) or (stimulus.shape[-2:] != stimulus_size[-2:]):
        raise ValueError(
            f"The spatial dimensions of stimulus {stimulus_name} did not match the "
            f"provided stimulus size: {stimulus.shape=}, provided_size: {stimulus_size=}"
        )
    return True


def load_stimuli(
    base_data_path: str,
    test_names: Iterable[str],
    stimulus_size: list[int],
    norm_mean: float | None,
    norm_std: float | None,
) -> dict[str, MoviesTrainTestSplit]:
    if not os.path.isdir(base_data_path):
        raise ValueError(f"{base_data_path=} is not a directory.")
    if len(stimulus_size) != 3:
        raise ValueError(f"Stimulus size should contain the channels, height and width, but was {stimulus_size=}")
    stimulus_size_tuple = (stimulus_size[0], stimulus_size[1], stimulus_size[2])

    test_stimuli_names = sorted(test_names)

    name_to_stimulus = {}
    # first load all stimuli from stimuli folder
    stimulus_folder = os.path.join(base_data_path, _STIMULUS_FOLDER)
    if os.path.isdir(stimulus_folder):
        for file_name in [x for x in os.listdir(stimulus_folder) if x.endswith(".npy")]:
            name = file_name.removesuffix(".npy")
            stim_path = os.path.join(stimulus_folder, file_name)
            stim = np.load(stim_path)
            _check_stimulus_size(stim, stimulus_size_tuple, stim_path)
            name_to_stimulus[name] = stim
    else:
        LOGGER.warning(f"Did not find {stimulus_folder=}")

    result = {}
    stimulus_processor = TrainTestStimuliProcessor(test_stimuli_names, name_to_stimulus, norm_mean, norm_std)
    # build MoviesTrainTestSplit for each session
    for file_name in [x for x in os.listdir(base_data_path) if x.endswith(".h5") or x.endswith(".hdf5")]:
        h5_path = os.path.join(base_data_path, file_name)
        with h5py.File(h5_path, "r") as f:
            stimuli_with_responses = sorted(
                x.removeprefix(_RESPONSES_PREFIX) for x in f.keys() if x.startswith(_RESPONSES_PREFIX)
            )
            if len(stimuli_with_responses) == 0:
                LOGGER.warning(f"No responses found for {file_name}: {list(f.keys())}")
                continue

            session_specific_stimuli = {
                x.removeprefix(_SESSION_SPECIFIC_STIMULUS_PREFIX): f[x]
                for x in f.keys()
                if (
                    x.startswith(_SESSION_SPECIFIC_STIMULUS_PREFIX)
                    and _check_stimulus_size(f[x], stimulus_size_tuple, f"{h5_path}/{x}")
                )
            }
            train_test_split = stimulus_processor.process(stimuli_with_responses, session_specific_stimuli)
        session_name = Path(file_name).stem
        result[session_name] = train_test_split
    return result


def load_responses(base_data_path: str, test_names: Iterable[str]) -> dict[str, ResponsesTrainTestSplit]:
    result = {}
    test_names_set = set(test_names)
    # build MoviesTrainTestSplit for each session
    for file_name in [x for x in os.listdir(base_data_path) if x.endswith(".h5") or x.endswith(".hdf5")]:
        with h5py.File(os.path.join(base_data_path, file_name), "r") as f:
            stimuli_with_responses = sorted(x for x in f.keys() if x.startswith(_RESPONSES_PREFIX))
            train_responses = np.concatenate(
                [f[x] for x in stimuli_with_responses if x.removeprefix(_RESPONSES_PREFIX) not in test_names_set],
                axis=-1,
            )
            test_responses_dict = {
                x.removeprefix(_RESPONSES_PREFIX): np.array(f[x])
                for x in stimuli_with_responses
                if x.removeprefix(_RESPONSES_PREFIX) in test_names_set
            }

            test_responses_neurons = {x.shape[0] for x in test_responses_dict.values()}
            if len(test_responses_neurons | {train_responses.shape[0]}) > 1:
                raise ValueError(
                    f"Train responses and test responses have a different number of neurons: "
                    f"{train_responses.shape[0]=} {test_responses_neurons=}"
                )
            # potentially improve: not sure if that fails for strings or other datatypes
            session_kwargs = {k: np.array(v) for k, v in f.get(_SESSION_INFO_KEY, {}).items()}

        session_name = Path(file_name).stem
        result[session_name] = ResponsesTrainTestSplit(
            train_responses, test_responses_dict, session_kwargs=session_kwargs
        )
    return result



import warnings
from copy import deepcopy
from typing import Literal, Optional, no_type_check

import numpy as np
import torch
from jaxtyping import Float
from tqdm.auto import tqdm

from openretina.data_io.base import ResponsesTrainTestSplit
from openretina.data_io.hoefling_2024.constants import CLIP_LENGTH, NUM_CLIPS, STIMULI_IDS


class NeuronDataSplitHoefling:
    def __init__(
        self,
        neural_responses: ResponsesTrainTestSplit,
        val_clip_idx: Optional[list[int]],
        num_clips: Optional[int],
        clip_length: Optional[int],
        roi_mask: Optional[Float[np.ndarray, "64 64"]] = None,
        roi_ids: Optional[Float[np.ndarray, " n_neurons"]] = None,
        scan_sequence_idx: Optional[int] = None,
        random_sequences: Optional[Float[np.ndarray, "n_clips n_sequences"]] = None,
        eye: Optional[Literal["left", "right"]] = None,
        group_assignment: Optional[Float[np.ndarray, " n_neurons"]] = None,
        key: Optional[dict] = None,
        use_base_sequence: Optional[bool] = False,
        **kwargs,
    ):
        """
        Initialize the NeuronData object.
        Boilerplate class to store neuron data train/test/validation splits before feeding into a dataloader.
        Customized for compatibility with the data format in Hoefling et al., 2024.

        Args:
            eye (str): The eye from which the neuron data is recorded.
            group_assignment (Float[np.ndarray, "n_neurons"]): The group assignment of neurons.
            key (dict): The key information for the neuron data,
                        includes date, exp_num, experimenter, field_id, stim_id.
            responses_final (Float[np.ndarray, "n_neurons n_timepoints"]): The responses of neurons.
            roi_coords (Float[np.ndarray, "n_neurons 2"]): The coordinates of regions of interest (ROIs).
            roi_ids (Float[np.ndarray, "n_neurons"]): The IDs of regions of interest (ROIs).
            scan_sequence_idx (int): The index of the scan sequence.
            stim_id (int): The ID of the stimulus. 5 is mouse natural scenes.
            random_sequences (Float[np.ndarray, "n_clips n_sequences"]): The random sequences of clips.
            val_clip_idx (List[int]): The indices of validation clips.
            num_clips (int): The number of clips.
            clip_length (int): The length of each clip.
            use_base_sequence (bool): Whether to re-order all training responses to use the same "base" sequence.
        """
        self.neural_responses = neural_responses

        self.num_neurons = self.neural_responses.n_neurons

        self.eye = eye if eye is not None else "right"
        self.group_assignment = group_assignment
        self.key = key
        self.roi_ids = roi_ids
        self.roi_coords = (
            np.array(self.transform_roi_mask(roi_mask), dtype=np.float32) if roi_mask is not None else None
        )
        self.scan_sequence_idx = scan_sequence_idx
        self.stim_id = self.neural_responses.stim_id
        self.clip_length = clip_length
        self.num_clips = num_clips
        self.random_sequences = random_sequences if random_sequences is not None else np.array([])
        self.use_base_sequence = use_base_sequence
        self.val_clip_idx = val_clip_idx

        self.responses_train_and_val, self.responses_test, self.test_responses_by_trial = (
            self.neural_responses.train.T,
            self.neural_responses.test_response.T,
            self.neural_responses.test_by_trial.transpose(0, 2, 1)
            if self.neural_responses.test_by_trial is not None
            else None,
        )
        self.responses_train, self.responses_val = self.compute_validation_responses()

    @property
    def response_dict(self):
        return {
            "train": torch.tensor(self.responses_train).to(torch.float),
            "validation": torch.tensor(self.responses_val).to(torch.float),
            "test": {
                "avg": torch.tensor(self.responses_test).to(torch.float),
                "by_trial": torch.tensor(self.test_responses_by_trial),
            },
        }

    @no_type_check
    def compute_validation_responses(self) -> tuple[np.ndarray, np.ndarray]:
        if self.stim_id in ["mb", "chirp"]:
            # Chirp and moving bar
            responses_val = np.array(np.nan)
            responses_train = self.responses_train_and_val
        else:
            movie_ordering = (
                np.arange(self.num_clips)
                if (len(self.random_sequences) == 0 or self.scan_sequence_idx is None)
                else self.random_sequences[:, self.scan_sequence_idx]
            )

            # Initialise validation responses

            base_movie_sorting = np.argsort(movie_ordering)

            validation_mask = np.ones_like(self.responses_train_and_val, dtype=bool)
            responses_val = np.zeros([len(self.val_clip_idx) * self.clip_length, self.num_neurons])

            # Compute validation responses and remove sections from training responses

            for i, ind1 in enumerate(self.val_clip_idx):
                grab_index = base_movie_sorting[ind1]
                responses_val[i * self.clip_length : (i + 1) * self.clip_length, :] = self.responses_train_and_val[
                    grab_index * self.clip_length : (grab_index + 1) * self.clip_length,
                    :,
                ]
                validation_mask[
                    (grab_index * self.clip_length) : (grab_index + 1) * self.clip_length,
                    :,
                ] = False

            if self.use_base_sequence:
                # Reorder training responses to use the same "base" sequence, which follows the numbering of clips.
                # This way all training responses are wrt the same order of clips, which can be useful in analysis.
                train_clip_idx = [i for i in range(self.num_clips) if i not in self.val_clip_idx]
                responses_train = np.zeros([len(train_clip_idx) * self.clip_length, self.num_neurons])
                for i, train_idx in enumerate(train_clip_idx):
                    grab_index = base_movie_sorting[train_idx]
                    responses_train[i * self.clip_length : (i + 1) * self.clip_length, :] = (
                        self.responses_train_and_val[
                            grab_index * self.clip_length : (grab_index + 1) * self.clip_length,
                            :,
                        ]
                    )
            else:
                responses_train = self.responses_train_and_val[validation_mask].reshape(-1, self.num_neurons)

        return responses_train, responses_val

    def transform_roi_mask(self, roi_mask):
        roi_coords = np.zeros((len(self.roi_ids), 2))
        for i, roi_id in enumerate(self.roi_ids):
            single_roi_mask = np.zeros_like(roi_mask)
            single_roi_mask[roi_mask == -roi_id] = 1
            roi_coords[i] = self.roi2readout(single_roi_mask)
        return roi_coords

    def roi2readout(
        self,
        single_roi_mask,
        roi_mask_pixelsize=2,
        readout_mask_pixelsize=50,
        x_offset=2.75,
        y_offset=2.75,
    ):
        """
        Maps a roi mask of a single roi from recording coordinates to model
        readout coordinates
        :param single_roi_mask: 2d array with nonzero values indicating the pixels
                of the current roi
        :param roi_mask_pixelsize: size of a pixel in the roi mask in um
        :param readout_mask_pixelsize: size of a pixel in the readout mask in um
        :param x_offset: x offset indicating the start of the recording field in readout mask
        :param y_offset: y offset indicating the start of the recording field in readout mask
        :return:
        """
        pixel_factor = readout_mask_pixelsize / roi_mask_pixelsize
        y, x = np.nonzero(single_roi_mask)
        y_trans, x_trans = y / pixel_factor, x / pixel_factor
        y_trans += y_offset
        x_trans += x_offset
        x_trans = x_trans.mean()
        y_trans = y_trans.mean()
        coords = np.asarray(
            [
                self.map_to_range(max=8, val=y_trans),
                self.map_to_range(max=8, val=x_trans),
            ],
            dtype=np.float32,
        )
        return coords

    def map_to_range(self, max, val):
        val = val / max
        val = val - 0.5
        val = val * 2
        return val


def compute_test_responses(
    neural_responses: Float[np.ndarray, "n_neurons time"],
    clip_length=CLIP_LENGTH,
    num_clips=NUM_CLIPS,
    response_type: Literal["natural", "chirp", "mb"] = "natural",
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    if response_type in ["chirp", "mb"]:
        # Chirp and moving bar
        responses_test = np.array(np.nan)
        test_responses_by_trial = [np.nan]
        responses_train_and_val = neural_responses.T

    else:
        assert clip_length is not None, "Clip length must be provided for natural scenes"
        assert num_clips is not None, "Number of clips must be provided for natural scenes"

        num_neurons = neural_responses.shape[0]

        responses_test = np.zeros((5 * clip_length, num_neurons))
        responses_train_and_val = np.zeros((num_clips * clip_length, num_neurons))

        test_responses_by_trial = []

        # Note: the hardcoded indices are the location of test clips in Hoefling 2024
        for idx in range(num_neurons):
            tmp = np.vstack(
                (
                    neural_responses[idx, : 5 * clip_length],
                    neural_responses[idx, 59 * clip_length : 64 * clip_length],
                    neural_responses[idx, 118 * clip_length :],
                )
            )
            test_responses_by_trial.append(tmp.T)  # type: ignore
            responses_test[:, idx] = np.mean(tmp, 0)
            responses_train_and_val[:, idx] = np.concatenate(
                (
                    neural_responses[idx, 5 * clip_length : 59 * clip_length],
                    neural_responses[idx, 64 * clip_length : 118 * clip_length],
                )
            )

    return responses_train_and_val.T, responses_test.T, np.asarray(test_responses_by_trial)


def upsample_traces(
    triggertimes,
    traces,
    tracestimes,
    stim_id: int,
    target_fr: int = 30,
    norm_by_std: bool = True,
) -> np.ndarray:
    """
    Upsamples the traces based on the stimulus type.

    Args:
        triggertimes (list): List of trigger times.
        traces (list): List of traces.
        tracestimes (list): List of trace times.
        stim_id (int): Stimulus ID.
        target_fr (int, optional): Target frame rate for upsampling. Default is 30.

    Returns:
        numpy.ndarray: Upsampled responses.

    Raises:
        NotImplementedError: If the stimulus ID is not implemented.
    """
    if stim_id == 1:
        # Chirp: each chirp has two triggers, one at the start and one 5s later, after a 2s OFF and 3s full field ON.
        # We need only the first trigger of each chirp for the upsampling.
        # 32.98999999 is the total chirp duration in seconds. Should be 33 but there is a small discrepancy
        chirp_starts = triggertimes[::2]
        upsampled_triggertimes = _upsample_triggertimes(32.98999999, 33, chirp_starts, target_fr)
    elif stim_id == 2:
        # Moving bar: each bar has one trigger at the start of the bar stim. Bar duration is 4s.
        # It is a bit more because each trigger has a duration of 3 frames at 60Hz, so around 50 ms.
        upsampled_triggertimes = _upsample_triggertimes(4.054001, 4.1, triggertimes, target_fr)
    elif stim_id == 5:
        # Movie stimulus
        # 4.966666 is the time between triggers in the movie stimulus.
        # It is not exactly 5s because it is not a perfect world :)
        upsampled_triggertimes = _upsample_triggertimes(4.9666667, 5, triggertimes, target_fr)
    else:
        raise NotImplementedError(f"Stimulus ID {stim_id} not implemented")

    upsampled_responses = np.zeros((traces.shape[0], len(upsampled_triggertimes)))
    for i in range(traces.shape[0]):
        upsampled_responses[i] = np.interp(upsampled_triggertimes, tracestimes[i].ravel(), traces[i].ravel())

    if norm_by_std:
        upsampled_responses = upsampled_responses / np.std(
            upsampled_responses, axis=1, keepdims=True
        )  # normalize response std

    return upsampled_responses


def _upsample_triggertimes(stim_empirical_duration, stim_theoretical_duration, triggertimes, target_fr):
    # upsample triggertimes to get 1 trigger per frame, (instead of just 1 trigger at the start of the sequence)
    upsampled_triggertimes = [
        np.linspace(t, t + stim_empirical_duration, round(stim_theoretical_duration * target_fr)) for t in triggertimes
    ]
    upsampled_triggertimes = np.concatenate(upsampled_triggertimes)

    return upsampled_triggertimes


def _apply_mask_to_field(data_dict: dict[str, dict], field: str, mask: Float[np.ndarray, " n_neurons"]) -> None:
    """
    Apply a mask to a specific field in a data dictionary.

    Args:
        data_dict (dict): A dictionary containing data fields.
        field (str): The field in the data dictionary to apply the mask to.
        mask (np.ndarray): The mask (assumed over neurons) to apply to each entry of the field.

    Returns:
        None

    Raises:
        IndexError: If the mask index is out of bounds for the field data.

    Examples:
        _apply_mask_to_field(data_dict, 'field_name', mask)"""

    for key in data_dict[field].keys():
        if key in ["roi_mask", "roi_coords"]:
            continue
        if isinstance(data_dict[field][key], np.ndarray) and len(data_dict[field][key]) > 0:
            if len(data_dict[field][key].shape) == 1:
                data_dict[field][key] = data_dict[field][key][mask]
            elif len(data_dict[field][key].shape) == 2:
                if data_dict[field][key].shape[0] == mask.shape[0]:
                    # If neurons are in the first dimension of the target array
                    data_dict[field][key] = data_dict[field][key][mask, :]
                else:
                    # If neurons are in the second dimension of the target array
                    data_dict[field][key] = data_dict[field][key][:, mask]
            else:
                raise IndexError(f"Index out of bounds for field {field} and key {key}.")


def _apply_qi_mask(data_dict, qi_types: list[str], qi_thresholds: list[float], logic="or"):
    """
    Applies quality thresholds as a mask to the data dictionary.

    Args:
        data_dict (dict): The data dictionary.
        qi_types (list): List of quality index types. Supported types are 'd' and 'chirp', corresponding to the
                        quality indices for the direction selectivity and chirp responses, respectively.
        qi_thresholds (list): List of quality index thresholds.
        logic (str): The logic to combine different qi_types. Can be 'and' or 'or'. Default is 'and'.

    Returns:
        dict: The updated data dictionary.
    """
    new_data_dict = deepcopy(data_dict)

    if logic not in ["and", "or"]:
        raise ValueError("logic must be either 'and' or 'or'")

    for field in new_data_dict.keys():
        masks = [
            new_data_dict[field][f"{qi_type}_qi"] >= qi_threshold
            for qi_type, qi_threshold in zip(qi_types, qi_thresholds, strict=True)
        ]

        if logic == "and":
            combined_mask = np.logical_and.reduce(masks)
        else:  # logic == 'or'
            combined_mask = np.logical_or.reduce(masks)

        _apply_mask_to_field(new_data_dict, field, combined_mask)

    return _clean_up_empty_fields(new_data_dict)


def _clean_up_empty_fields(data_dict, check_field="group_assignment"):
    """
    Remove empty fields from the data dictionary.

    Args:
        data_dict (dict): The data dictionary.
        check_field (str, optional): The field to check for emptiness. Defaults to "group_assignment".

    Returns:
        dict: The updated data dictionary.
    """
    return {k: v for k, v in data_dict.items() if len(v[check_field]) > 0}


def _mask_by_cell_type(data_dict, cell_types: list[int] | int):
    if isinstance(cell_types, int):
        cell_types = [cell_types]

    new_data_dict = deepcopy(data_dict)
    for field in new_data_dict.keys():
        mask = np.isin(new_data_dict[field]["group_assignment"], cell_types)
        _apply_mask_to_field(new_data_dict, field, mask)

    return _clean_up_empty_fields(new_data_dict)


def _mask_by_classifier_confidence(data_dict, min_confidence: float):
    new_data_dict = deepcopy(data_dict)
    for field in new_data_dict.keys():
        mask = new_data_dict[field]["group_confidences"].max(axis=1) >= min_confidence
        _apply_mask_to_field(new_data_dict, field, mask)

    return _clean_up_empty_fields(new_data_dict)


def make_final_responses(
    data_dict: dict,
    response_type: Literal["natural", "chirp", "mb"] = "natural",
    trace_type: Literal["spikes", "raw", "preprocessed", "detrended"] = "spikes",
    d_qi: Optional[float] = None,
    chirp_qi: Optional[float] = None,
    qi_logic: Literal["and", "or"] = "or",
    scale_traces: float = 1.0,
    norm_by_std: bool = True,
):
    upsampled_data_dict = upsample_all_responses(
        data_dict, response_type, trace_type, d_qi, chirp_qi, qi_logic, scale_traces, norm_by_std
    )

    responses_dictionary = {}

    for field in upsampled_data_dict.keys():
        train_responses, test_responses, test_responses_by_trial = compute_test_responses(
            upsampled_data_dict[field]["responses_final"],
            response_type=response_type,
        )
        responses_dictionary[field] = ResponsesTrainTestSplit(
            train=train_responses,
            test=test_responses,
            test_by_trial=test_responses_by_trial,
            stim_id=response_type,
            session_kwargs={
                "eye": upsampled_data_dict[field]["eye"],
                "scan_sequence_idx": upsampled_data_dict[field]["scan_sequence_idx"],
                "group_assignment": upsampled_data_dict[field]["group_assignment"],
                "roi_ids": upsampled_data_dict[field]["roi_ids"],
                "roi_mask": upsampled_data_dict[field]["roi_mask"],
            },
        )

    return responses_dictionary


def upsample_all_responses(
    data_dict: dict,
    response_type: Literal["natural", "chirp", "mb"] = "natural",
    trace_type: Literal["spikes", "raw", "preprocessed", "detrended"] = "spikes",
    d_qi: Optional[float] = None,
    chirp_qi: Optional[float] = None,
    qi_logic: Literal["and", "or"] = "or",
    scale_traces: float = 1.0,
    norm_by_std: bool = True,
):
    """
    Converts inferred spikes into final responses by upsampling the traces of all sessions of a given response_type.
    This is to match the framerate used in the stimulus presentation.

    Args:
        data_dict (dict): A dictionary containing the data.
        response_type (str, optional): The type of response. Defaults to "natural".

    Returns:
        dict: The updated data dictionary with final responses.
    Raises:
        NotImplementedError: If the conversion is not yet implemented for the given response type.
    """

    new_data_dict = deepcopy(data_dict)

    stim_id = STIMULI_IDS.get(response_type, None)
    if stim_id is None:
        raise NotImplementedError(f"Conversion not yet implemented for response type {response_type}")

    for field in tqdm(
        new_data_dict.keys(),
        desc=f"Upsampling {response_type} {trace_type} traces to get final responses.",
    ):
        if trace_type == "detrended":
            raw_traces = new_data_dict[field][f"{response_type}_raw_traces"]
            smoothed_traces = new_data_dict[field][f"{response_type}_smoothed_traces"]
            traces = raw_traces - smoothed_traces
        elif trace_type == "spikes":
            try:
                traces = new_data_dict[field][f"{response_type}_inferred_spikes"]
            except KeyError:
                # For new data format
                traces = new_data_dict[field][f"{response_type}_spikes"]
        else:
            traces = new_data_dict[field][f"{response_type}_{trace_type}_traces"]

        triggertimes = new_data_dict[field][f"{response_type}_trigger_times"][0]

        try:
            tracestimes = new_data_dict[field][f"{response_type}_traces_times"]
        except KeyError:
            # New djimaing exports have a different save format for trace_times
            traces_t0 = np.tile(
                np.atleast_1d(new_data_dict[field][f"{response_type}_traces_t0"].squeeze())[:, None],
                (1, traces.shape[1]),
            )
            traces_dt = np.tile(
                np.atleast_1d(new_data_dict[field][f"{response_type}_traces_dt"].squeeze())[:, None],
                (1, traces.shape[1]),
            )
            tracestimes = np.tile(np.arange(traces.shape[1]), reps=(traces.shape[0], 1)) * traces_dt + traces_t0

        upsampled_traces = (
            upsample_traces(
                triggertimes=triggertimes,
                traces=traces,
                tracestimes=tracestimes,
                stim_id=stim_id,
                norm_by_std=norm_by_std,
            )
            * scale_traces
        )

        new_data_dict[field][f"{response_type}_responses_final"] = upsampled_traces

        if "responses_final" in new_data_dict[field]:
            warnings.warn(
                f"You seem to already have computed `responses_final` for a "
                f"stim_id of {new_data_dict[field]['stim_id']}. "
                f"Overwriting with {stim_id} ({response_type})."
            )
        new_data_dict[field]["responses_final"] = upsampled_traces
        new_data_dict[field]["stim_id"] = stim_id

    d_qi = d_qi if d_qi is not None else 0.0
    chirp_qi = chirp_qi if chirp_qi is not None else 0.0
    # Apply quality indices mask only if requested
    if d_qi > 0.0 or chirp_qi > 0.0:
        new_data_dict = _apply_qi_mask(new_data_dict, ["d", "chirp"], [d_qi, chirp_qi], qi_logic)

    return new_data_dict


def filter_responses(
    all_responses: dict[str, dict],
    filter_cell_types: bool = False,
    cell_types_list: Optional[list[int] | int] = None,
    chirp_qi: float = 0.35,
    d_qi: float = 0.6,
    qi_logic: Literal["and", "or"] = "or",
    filter_counts: bool = True,
    count_threshold: int = 10,
    classifier_confidence: float = 0.25,
    verbose: bool = False,
) -> dict[str, dict]:
    """
    This function processes the input dictionary of neuron responses, applying various filters
    to exclude unwanted data based on the provided parameters. It can filter by cell types,
    quality indices, classifier confidence, and the number of responding cells, while also
    providing verbose output for tracking the filtering process.
    Note: default arguments are from the Hoefling et al., 2024 paper.

    Args:
        all_responses (Dict[str, dict]): A dictionary containing neuron response data.
        filter_cell_types (bool, optional): Whether to filter by specific cell types. Defaults to False.
        cell_types_list (Optional[List[int] | int], optional): List or single value of cell types to filter.
                                                                Defaults to None.
        chirp_qi (float, optional): Quality index threshold for chirp responses. Defaults to 0.35.
        d_qi (float, optional): Quality index threshold for d responses. Defaults to 0.6.
        qi_logic (Literal["and", "or"], optional): The logic to combine different quality indices. Defaults to "and".
        filter_counts (bool, optional): Whether to filter based on response counts. Defaults to True.
        count_threshold (int, optional): Minimum number of responding cells required. Defaults to 10.
        classifier_confidence (float, optional): Minimum confidence level for classifier responses. Defaults to 0.3.
        verbose (bool, optional): If True, prints detailed filtering information. Defaults to False.

    Returns:
        Dict[str, dict]: A filtered dictionary of neuron responses that meet the specified criteria.
    """

    def print_verbose(message):
        if verbose:
            print(message)

    def get_n_neurons(all_responses) -> int:
        return sum(len(field["group_assignment"]) for field in all_responses.values())

    original_neuron_count = get_n_neurons(all_responses)
    print_verbose(f"Original dataset contains {original_neuron_count} neurons over {len(all_responses)} fields")
    print_verbose(" ------------------------------------ ")

    # Filter by cell types
    if filter_cell_types and cell_types_list is not None:
        all_rgcs_responses_ct_filtered = _mask_by_cell_type(all_responses, cell_types_list)
        print_verbose(
            f"Dropped {len(all_responses) - len(all_rgcs_responses_ct_filtered)} fields that did not "
            f"contain the target cell types ({len(all_rgcs_responses_ct_filtered)} remaining)"
        )
        dropped_n_cell_types = original_neuron_count - get_n_neurons(all_rgcs_responses_ct_filtered)
        print_verbose(
            f"Overall, dropped {dropped_n_cell_types} neurons of non-target cell types "
            f"(-{(dropped_n_cell_types) / original_neuron_count:.2%})."
        )
        print_verbose(" ------------------------------------ ")
    else:
        all_rgcs_responses_ct_filtered = deepcopy(all_responses)

    count_before_checks = get_n_neurons(all_rgcs_responses_ct_filtered)

    # Apply quality checks
    all_rgcs_responses_ct_filtered = _apply_qi_mask(
        all_rgcs_responses_ct_filtered, ["d", "chirp"], [d_qi, chirp_qi], qi_logic
    )

    dropped_n_qi = count_before_checks - get_n_neurons(all_rgcs_responses_ct_filtered)

    print_verbose(
        f"Dropped {len(all_responses) - len(all_rgcs_responses_ct_filtered)} fields with quality indices "
        f"below threshold ({len(all_rgcs_responses_ct_filtered)} remaining)"
    )
    print_verbose(
        f"Overall, dropped {dropped_n_qi} neurons over quality checks (-{dropped_n_qi / count_before_checks:.2%})."
    )
    print_verbose(" ------------------------------------ ")

    # Filter by classifier confidence
    if classifier_confidence is not None and classifier_confidence > 0.0:
        all_rgcs_responses_confidence_filtered = _mask_by_classifier_confidence(
            all_rgcs_responses_ct_filtered, classifier_confidence
        )
        dropped_n_classifier = get_n_neurons(all_rgcs_responses_ct_filtered) - get_n_neurons(
            all_rgcs_responses_confidence_filtered
        )
        print_verbose(
            f"Dropped {len(all_rgcs_responses_ct_filtered) - len(all_rgcs_responses_confidence_filtered)} fields with "
            f"classifier confidences below {classifier_confidence}"
        )
        print_verbose(
            f"Overall, dropped {dropped_n_classifier} neurons with classifier confidences below {classifier_confidence}"
            f" (-{dropped_n_classifier / get_n_neurons(all_rgcs_responses_ct_filtered):.2%})."
        )
        print_verbose(" ------------------------------------ ")
    else:
        all_rgcs_responses_confidence_filtered = all_rgcs_responses_ct_filtered

    # Filter by low counts
    if filter_counts:
        all_rgcs_responses = {
            k: v
            for k, v in all_rgcs_responses_confidence_filtered.items()
            if len(v["group_assignment"]) > count_threshold
        }
        dropped_n_counts = get_n_neurons(all_rgcs_responses_confidence_filtered) - get_n_neurons(all_rgcs_responses)
        print_verbose(
            f"Dropped {len(all_rgcs_responses_confidence_filtered) - len(all_rgcs_responses)} fields with less than "
            f"{count_threshold} responding cells ({len(all_rgcs_responses)} remaining)"
        )
        print_verbose(
            f"Overall, dropped {dropped_n_counts} neurons in fields with less than {count_threshold} responding cells "
            f"(-{dropped_n_counts / get_n_neurons(all_rgcs_responses_confidence_filtered):.2%})."
        )
    else:
        all_rgcs_responses = all_rgcs_responses_confidence_filtered

    print_verbose(" ------------------------------------ ")
    print_verbose(
        f"Final dataset contains {get_n_neurons(all_rgcs_responses)} neurons over {len(all_rgcs_responses)} fields"
    )
    final_n_dropped = original_neuron_count - get_n_neurons(all_rgcs_responses)
    print(f"Total number of cells dropped: {final_n_dropped} (-{(final_n_dropped) / original_neuron_count:.2%})")

    # Clean up RAM
    del all_rgcs_responses_ct_filtered
    del all_rgcs_responses_confidence_filtered

    return all_rgcs_responses



from collections.abc import Iterable
from typing import Optional

import einops
import torch
import torch.nn.functional as F
from jaxtyping import Float


def _gaussian_1d_kernel(sigma: float, kernel_size: int) -> torch.Tensor:
    """Create a 1D Gaussian kernel."""
    x = torch.arange(kernel_size).float() - kernel_size // 2
    kernel = torch.exp(-(x**2) / (2 * sigma**2))
    kernel = kernel / kernel.sum()  # Normalize to ensure the sum is 1
    return kernel


class StimulusRegularizationLoss:
    def forward(self, stimulus: torch.Tensor) -> torch.Tensor:
        return 0.0  # type: ignore


class RangeRegularizationLoss(StimulusRegularizationLoss):
    def __init__(
        self,
        min_max_values: Iterable[tuple[float | None, float | None]],
        max_norm: float | None,
        factor: float = 1.0,
    ):
        self._min_max_values = list(min_max_values)
        self._max_norm = max_norm
        self._factor = factor

    def forward(self, stimulus: torch.Tensor) -> torch.Tensor:
        """Penalizes the stimulus if it is outside the range defined by min_max_values."""
        loss: torch.Tensor = 0.0  # type: ignore
        for i, (min_val, max_val) in enumerate(self._min_max_values):
            stimulus_i = stimulus[:, i]
            if min_val is not None:
                loss += torch.sum(torch.relu(min_val - stimulus_i))
            if max_val is not None:
                loss += torch.sum(torch.relu(stimulus_i - max_val))

        if self._max_norm is not None:
            # Add a loss such that the norm of the stimulus is lower than max_norm
            norm_penalty = torch.relu(torch.norm(stimulus) - self._max_norm)
            loss += norm_penalty

        loss *= self._factor
        return loss


class StimulusPostprocessor:
    """Base class for stimulus clippers."""

    def process(self, x: torch.Tensor) -> torch.Tensor:
        """x.shape: batch x channels x time x n_rows x n_cols"""
        return x


class ChangeNormJointlyClipRangeSeparately(StimulusPostprocessor):
    """First change the norm and afterward clip the value of x to some specified range"""

    def __init__(
        self,
        min_max_values: Iterable[tuple[Optional[float], Optional[float]]],
        norm: float | None,
    ):
        self._norm = norm
        self._min_max_values = list(min_max_values)

    def process(self, x: torch.Tensor) -> torch.Tensor:
        assert x.shape[1] == len(self._min_max_values), (
            f"Expected {len(self._min_max_values)} channels in dim 1, got {x.shape=}"
        )

        if self._norm is not None:
            # Re-normalize
            x_norm = torch.linalg.vector_norm(x.view(len(x), -1), dim=-1)
            renorm = x * (self._norm / x_norm).view(len(x), *[1] * (x.dim() - 1))
        else:
            renorm = x

        # Clip
        clipped_array = []
        for i, (min_val, max_val) in enumerate(self._min_max_values):
            clipped = renorm[:, i]
            if min_val is not None or max_val is not None:
                clipped = torch.clamp(clipped, min=min_val, max=max_val)
            clipped_array.append(clipped)
        result = torch.stack(clipped_array, dim=1)

        return result

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._norm=}, {self._min_max_values=})"


class TemporalGaussianLowPassFilterProcessor(StimulusPostprocessor):
    """Uses a 1d Gaussian filter to convolve the stimulus over the temporal dimension.
    This acts as a low pass filter."""

    def __init__(
        self,
        sigma: float,
        kernel_size: int,
        device: str = "cpu",
    ):
        kernel = _gaussian_1d_kernel(sigma, kernel_size)
        self._kernel = kernel.to(device)

    def process(self, x: Float[torch.Tensor, "batch_dim channels time height width"]) -> torch.Tensor:
        """
        Apply a Gaussian low-pass filter to the stimulus tensor along the temporal dimension.

        Arguments:
            x (Tensor): Tensor of shape (batch_dim, channels, time_dim, height, width)
        Returns:
            Tensor: The filtered stimulus tensor.
        """
        # Create the Gaussian kernel in the temporal dimension
        kernel = einops.repeat(self._kernel.to(x.device), "s -> c 1 s 1 1", c=x.shape[1])

        # Apply convolution in the temporal dimension (axis 2)
        # We need to ensure that the kernel is convolved only along the time dimension.
        filtered_stimulus = F.conv3d(x, kernel, padding="same", groups=x.shape[1])

        return filtered_stimulus



import os
import tempfile
from pathlib import Path

import pytest

from openretina.utils.file_utils import get_file_size, get_local_file_path, is_target_present


@pytest.mark.parametrize(
    "possibly_remote_file_path",
    [
        # single file
        "https://huggingface.co/datasets/open-retina/open-retina/"
        "blob/main/euler_lab/hoefling_2024/responses/"
        "rgc_natsim_subset_only_naturalspikes_2024-08-14.h5",
    ],
)
def test_get_local_file_path(possibly_remote_file_path: str) -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        local_file_path = get_local_file_path(possibly_remote_file_path, cache_folder=temp_dir)
        assert is_target_present(Path(temp_dir), local_file_path)
        date = os.path.getmtime(local_file_path)
        # Call the function a second time to see if caching works
        local_file_path_two = get_local_file_path(possibly_remote_file_path, cache_folder=temp_dir)
        assert local_file_path == local_file_path_two
        date_two = os.path.getmtime(local_file_path)
        assert date == date_two


@pytest.mark.parametrize(
    "url",
    [
        # single file
        "https://huggingface.co/datasets/open-retina/open-retina/"
        "blob/main/euler_lab/hoefling_2024/responses/"
        "rgc_natsim_subset_only_naturalspikes_2024-08-14.h5",
        # zip file
        "https://huggingface.co/datasets/open-retina/open-retina/resolve"
        "/main/baccus_lab/maheswaranathan_2023/neural_code_data.zip",
    ],
)
def test_get_file_size(url: str) -> None:
    size = get_file_size(url)
    assert size > 0



"""
Models in this folder are under development. They can depend on the legacy codebase or on neuralpredictors.
"""

import torch
import torch.nn as nn
from neuralpredictors.layers.readouts import FullGaussian2d, Gaussian3d, MultiReadoutBase


class Encoder(nn.Module):
    """
    puts together all parts of model (core, readouts) and defines a forward
    function which will return the output of the model; PyTorch then allows
    to call .backward() on the Encoder which will compute the gradients
    """

    def __init__(
        self,
        core: nn.Module,
        readout,
    ):
        super().__init__()
        self.core = core
        self.readout = readout
        self.detach_core = False

    def forward(self, x: torch.Tensor, data_key: str | None = None, detach_core: bool = False, **kwargs):
        self.detach_core = detach_core
        x = self.core(x, data_key=data_key)
        if self.detach_core:
            x = x.detach()
        x = self.readout(x, data_key=data_key)
        return x

    def readout_keys(self) -> list[str]:
        return self.readout.readout_keys()


class MultiGaussian3d(MultiReadoutBase):
    def __init__(self, in_shape_dict, n_neurons_dict, **kwargs):
        super().__init__(in_shape_dict, n_neurons_dict, base_readout=Gaussian3d, **kwargs)

    def regularizer(self, data_key):
        return 0


class MultiGaussian2d(MultiReadoutBase):
    def __init__(self, in_shape_dict, n_neurons_dict, **kwargs):
        super().__init__(
            in_shape_dict,
            n_neurons_dict,
            base_readout=FullGaussian2d,
            gauss_type="full",
            **kwargs,
        )



"""
Minimal stimulus loading utilities to train a model on the data used in Maheswaranathan et al. 2023

Paper: https://doi.org/10.1016/j.neuron.2023.06.007
Data: https://doi.org/10.25740/rk663dm5577

OpenRetina provides a mirror of the dataset on huggingface:
https://huggingface.co/datasets/open-retina/open-retina/resolve/main/baccus_lab/maheswaranathan_2023/neural_code_data.zip
"""

import os
from typing import Literal

from openretina.data_io.base import MoviesTrainTestSplit, normalize_train_test_movies
from openretina.utils.file_utils import get_local_file_path
from openretina.utils.h5_handling import load_dataset_from_h5

CLIP_LENGTH = 90  # in frames @ 30 fps


def load_all_stimuli(
    base_data_path: str | os.PathLike,
    stim_type: Literal["naturalscene", "whitenoise"] = "naturalscene",
    normalize_stimuli: bool = True,
) -> dict[str, MoviesTrainTestSplit]:
    """
    Load all stimuli from sessions within subfolders in a given base data path.

    The base data path should point to the location of the `neural_code_data` folder.
    (See https://doi.org/10.25740/rk663dm5577 for dataset download).

    Alternatively, base_data_path can point directly to our huggingface mirror of the dataset, which will then
    be downloaded and extracted automatically to the openretina cache directory.
    https://huggingface.co/datasets/open-retina/open-retina/resolve/main/baccus_lab/maheswaranathan_2023/neural_code_data.zip
    """
    # Resolve data path
    base_data_path = get_local_file_path(str(base_data_path))
    full_data_path = os.path.join(base_data_path, "ganglion_cell_data")

    stimuli_all_sessions = {}
    for session in [x.name for x in os.scandir(os.fspath(full_data_path)) if x.is_dir()]:
        session_path = os.path.normpath(os.path.join(full_data_path, session))
        for recording_file in os.listdir(session_path):
            if str(recording_file).endswith(f"{stim_type}.h5"):
                recording_file = os.path.join(session_path, recording_file)

                print(f"Loading stimuli from {recording_file}")

                # Load video stimuli
                train_video = load_dataset_from_h5(recording_file, "/train/stimulus")
                test_video = load_dataset_from_h5(recording_file, "/test/stimulus")

                # Add channel dimension
                train_video = train_video[None, ...]
                test_video = test_video[None, ...]

                if normalize_stimuli:
                    train_video, test_video, norm_dict = normalize_train_test_movies(train_video, test_video)
                else:
                    norm_dict = {"norm_mean": None, "norm_std": None}

                stimuli_all_sessions[str(session)] = MoviesTrainTestSplit(
                    train=train_video,
                    test=test_video,
                    stim_id=stim_type,
                    random_sequences=None,
                    norm_mean=norm_dict["norm_mean"],
                    norm_std=norm_dict["norm_std"],
                )
    return stimuli_all_sessions



import pytest
from torch.utils.data import DataLoader, IterableDataset

from openretina.data_io.cyclers import LongCycler, ShortCycler


class Dataset(IterableDataset):
    def __init__(self, name: str, num_elements: int):
        self._name = name
        self._num_elements = num_elements

    def __iter__(self):
        for i in range(self._num_elements):
            yield f"{self._name}_{i}"

    def __len__(self) -> int:
        return self._num_elements


def get_dataloaders(num_elements_per_loader: list[int]) -> dict[str, DataLoader]:
    dataloaders = {}
    for i, num_elements in enumerate(num_elements_per_loader):
        name = f"dataset{i}"
        ds = Dataset(name, num_elements)
        dl = DataLoader(ds, drop_last=False)
        dataloaders[name] = dl
    return dataloaders


def get_all_elements_from_dataloaders(dataloaders) -> list:
    elements_list = [[x[0] for x in dl] for dl in dataloaders]
    elements = sum(elements_list, [])
    return elements


def get_all_elements_from_cycler_dataloader(dataloader) -> list:
    all_elements_lists = list(dataloader)
    all_elements = [x[1][0] for x in all_elements_lists]
    return all_elements


@pytest.mark.parametrize(
    "num_workers",
    [0, 1, 3, 4],
)
def test_short_cycler(num_workers: int):
    dataloaders = get_dataloaders([1, 20, 5, 10])

    all_elements_dataloader = sorted(get_all_elements_from_dataloaders(dataloaders.values()))
    cycler = ShortCycler(dataloaders)
    dl_cycler = DataLoader(cycler, num_workers=num_workers, batch_size=None, drop_last=False)
    all_elements_cycler = sorted(get_all_elements_from_cycler_dataloader(dl_cycler))
    assert all_elements_cycler == all_elements_dataloader


@pytest.mark.parametrize(
    "num_workers, shuffle",
    [
        (0, False),
        (0, True),
        # (1, False),
        # (1, True),
        # (4, False),
        # (4, True),
    ],
)
def test_long_cycler(num_workers: int, shuffle: bool):
    dataloaders = get_dataloaders([1, 20, 5, 1])

    all_elements_dataloader = get_all_elements_from_dataloaders(dataloaders.values())
    cycler = LongCycler(dataloaders, shuffle=shuffle)
    dl_cycler = DataLoader(cycler, num_workers=num_workers, batch_size=None, drop_last=False)
    all_elements_cycler = get_all_elements_from_cycler_dataloader(dl_cycler)

    # cycler can produce more elements if the individual dataloaders have different sizes
    assert len(all_elements_dataloader) <= len(all_elements_cycler)
    # cycler should produce all elements in dataloaders
    assert set(all_elements_dataloader) == set(all_elements_cycler)



from typing import Callable, TypeVar

import torch
from torch import Tensor

from openretina.insilico.stimulus_optimization.optimization_stopper import OptimizationStopper
from openretina.insilico.stimulus_optimization.regularizer import (
    StimulusPostprocessor,
    StimulusRegularizationLoss,
)

T = TypeVar("T")


def convert_to_list(x: list[T] | T | None) -> list[T]:
    if x is None:
        return []
    elif isinstance(x, list):
        return x
    else:
        return [x]


def optimize_stimulus(
    stimulus: Tensor,
    optimizer_init_fn: Callable[[list[torch.Tensor]], torch.optim.Optimizer],
    objective_object,
    optimization_stopper: OptimizationStopper,
    stimulus_regularization_loss: list[StimulusRegularizationLoss] | StimulusRegularizationLoss | None = None,
    stimulus_postprocessor: list[StimulusPostprocessor] | StimulusPostprocessor | None = None,
) -> None:
    """
    Optimize a stimulus to maximize a given objective while minimizing a regularizing function.
    The stimulus is modified in place.
    """
    optimizer = optimizer_init_fn([stimulus])

    for _ in range(optimization_stopper.max_iterations):
        objective = objective_object.forward(stimulus)
        # Maximizing the objective, minimizing the regularization loss
        loss = -objective
        for reg_loss_module in convert_to_list(stimulus_regularization_loss):
            regularization_loss = reg_loss_module.forward(stimulus)
            loss += regularization_loss

        optimizer.zero_grad()
        loss.backward()
        optimizer.step()
        for postprocessor in convert_to_list(stimulus_postprocessor):
            stimulus.data = postprocessor.process(stimulus.data)
        if optimization_stopper.early_stop(float(loss.item())):
            break
    stimulus.detach_()  # Detach the tensor from the computation graph



import os


def get_cache_directory() -> str:
    home_dir = os.path.expanduser("~")
    return os.getenv("OPENRETINA_CACHE_DIRECTORY", os.path.join(home_dir, "openretina_cache"))


os.environ["OPENRETINA_CACHE_DIRECTORY"] = get_cache_directory()



"""
Adapted from sinzlab/neuralpredictors/training/cyclers.py
"""

import math
import random

import torch.utils.data
from torch.utils.data import DataLoader


def cycle(iterable):
    """
    itertools.cycle without caching.
    See: https://github.com/pytorch/pytorch/issues/23900
    """
    iterator = iter(iterable)
    while True:
        try:
            yield next(iterator)
        except StopIteration:
            iterator = iter(iterable)


class LongCycler(torch.utils.data.IterableDataset):
    """
    Cycles through a dictionary of data loaders until the loader with the largest size is exhausted.
    In practice, takes one batch from each loader in each iteration.
    Necessary for dataloaders of unequal size.
    Note: iterable dataloaders as this one can lead to duplicate data when using multiprocessing.
    """

    def __init__(self, loaders: dict[str, DataLoader], shuffle: bool = True):
        self.loaders = loaders
        self.max_batches = max(len(loader) for loader in self.loaders.values())
        self.shuffle = shuffle

    def __iter__(self):
        worker_info = torch.utils.data.get_worker_info()
        if worker_info is not None and worker_info.num_workers > 1:
            raise NotImplementedError("LongCycler does not support multiple workers yet.")

        keys = sorted(self.loaders.keys())

        if self.shuffle:
            random.shuffle(keys)

        # Create cycles for each loader
        cycles = [cycle(self.loaders[k]) for k in keys]
        total_iterations = len(self.loaders) * self.max_batches

        # Yield batches in the assigned range
        for k, loader, _ in zip(cycle(keys), cycle(cycles), range(total_iterations)):
            yield k, next(loader)

    def __len__(self):
        return len(self.loaders) * self.max_batches


class ShortCycler(torch.utils.data.IterableDataset):
    """
    Cycles through the elements of each dataloader without repeating any element.
    """

    def __init__(self, loaders: dict[str, DataLoader]):
        self.loaders = loaders

    def _get_keys(self) -> list[str]:
        sorted_keys = sorted(self.loaders.keys())
        worker_info = torch.utils.data.get_worker_info()

        if worker_info is not None:
            if worker_info.num_workers > len(sorted_keys):
                raise ValueError(f"Too many workers for {len(sorted_keys)} sessions: {worker_info=}")

            sess_per_worker = math.ceil(len(sorted_keys) / worker_info.num_workers)
            start_idx = sess_per_worker * worker_info.id
            return sorted_keys[start_idx : start_idx + sess_per_worker]
        else:
            return sorted_keys

    def __iter__(self):
        for k in self._get_keys():
            for example in self.loaders[k]:
                yield k, example



import pytest
import torch

from openretina.models.core_readout import _MODEL_NAME_TO_REMOTE_LOCATION, load_core_readout_from_remote


@pytest.mark.parametrize("model_name", sorted(_MODEL_NAME_TO_REMOTE_LOCATION.keys()))
def test_load_core_readout_from_remote(model_name: str) -> None:
    num_batches = 1

    model = load_core_readout_from_remote(model_name, "cpu")
    stimulus = torch.rand(model.stimulus_shape(time_steps=50, num_batches=num_batches))
    responses = model.forward(stimulus)

    assert responses.shape[0] == num_batches



"""
Minimal stimuli loading utilities to train a model on the data used in Karamanlis et al. 2024

Paper: https://doi.org/10.1038/s41586-024-08212-3
Data: https://doi.org/10.12751/g-node.ejk8kx
"""

import os
import warnings
from pathlib import Path
from typing import Literal

import numpy as np
import torch
from einops import rearrange
from torchvision.transforms import Resize
from tqdm.auto import tqdm

from openretina.data_io.base import MoviesTrainTestSplit, normalize_train_test_movies
from openretina.utils.file_utils import get_local_file_path, unzip_and_cleanup
from openretina.utils.h5_handling import load_dataset_from_h5


def load_stimuli_for_session(
    session_path: str | os.PathLike,
    stim_type: Literal["fixationmovie", "frozencheckerflicker", "gratingflicker", "imagesequence"],
    downsampled_size: tuple[int, int],
    normalize_stimuli: bool,
) -> MoviesTrainTestSplit | None:
    """
    Load stimuli for a single session.

    Args:
        session_path (str): Path to the session directory.
        stim_type (str): The stimulus type to filter files.
        downsampled_size (tuple[int, int]): Size to downsample the stimuli.
        normalize_stimuli (bool): Whether to normalize the stimuli.

    Returns:
        MoviesTrainTestSplit | None: Loaded stimuli for the session or None if no relevant file found.
    """
    if str(session_path).endswith(".zip"):
        session_path = unzip_and_cleanup(Path(session_path))

    mat_file = None
    npz_file = None
    downsample = Resize(downsampled_size)

    for recording_file in os.listdir(session_path):
        full_path = os.path.join(session_path, recording_file)

        # First check for a `.npz` file, which is a saved version of the preprocessing of the stimuli
        if recording_file.endswith(
            f"{stim_type}_data_extracted_stimuli_{downsampled_size[0]}_{downsampled_size[1]}.npz"
        ):
            npz_file = full_path
            break  # Prefer `.npz` file

        # Fall back to the original `.mat` file if no `.npz` file is found (e.g., for the first run of the script)
        elif recording_file.endswith(f"{stim_type}_data.mat"):
            mat_file = full_path

    if npz_file:
        tqdm.write(f"Loading stimuli from cached {npz_file}")
        video_data = np.load(npz_file)
        train_video = video_data["train_data"]
        test_video = video_data["test_data"]

    elif mat_file:
        tqdm.write(f"Loading stimuli from {mat_file}")
        train_images = load_dataset_from_h5(mat_file, "runningImages")
        test_images = load_dataset_from_h5(mat_file, "frozenImages")

        frozen_fixations = load_dataset_from_h5(mat_file, "frozenfixations").astype(int)
        running_fixations = load_dataset_from_h5(mat_file, "runningfixations").astype(int)

        test_video = return_fix_movie_torch((600, 800), rearrange(test_images, "n x y -> y x n"), frozen_fixations.T)
        test_video = downsample(rearrange(test_video, "h w n -> 1 n h w")).cpu().numpy()

        train_videos = []
        for trial in tqdm(running_fixations, desc=f"Composing training video for {Path(mat_file).parent.name}"):
            train_snippet = return_fix_movie_torch((600, 800), rearrange(train_images, "n x y -> y x n"), trial.T)
            train_videos.append(downsample(rearrange(train_snippet, "h w n -> 1 n h w")))
        train_video = torch.cat(train_videos, dim=1).cpu().numpy()

        np.savez_compressed(
            f"{mat_file[:-4]}_extracted_stimuli_{downsampled_size[0]}_{downsampled_size[1]}.npz",
            train_data=train_video,
            test_data=test_video,
        )
    else:
        warnings.warn(
            f"No file with postfix {stim_type}_data.mat (or preprocessed version) found in folder {session_path}",
        )
        return None

    if normalize_stimuli:
        train_video, test_video, norm_dict = normalize_train_test_movies(train_video, test_video)
    else:
        norm_dict = {"norm_mean": None, "norm_std": None}

    return MoviesTrainTestSplit(
        train=train_video,
        test=test_video,
        stim_id=stim_type,
        random_sequences=None,
        norm_mean=norm_dict["norm_mean"],
        norm_std=norm_dict["norm_std"],
    )


def load_all_stimuli(
    base_data_path: str | os.PathLike,
    stim_type: Literal["fixationmovie", "frozencheckerflicker", "gratingflicker", "imagesequence"] = "fixationmovie",
    normalize_stimuli: bool = True,
    specie: Literal["mouse", "marmoset"] = "mouse",
    downsampled_size: tuple[int, int] = (60, 80),
) -> dict[str, MoviesTrainTestSplit]:
    """
    Load stimuli for all sessions.

    Args:
        base_data_path (str | os.PathLike): Base directory containing session data.
                                            Can also be the path to the "sessions" folder in the huggingface mirror.
        "https://huggingface.co/datasets/open-retina/open-retina/tree/main/gollisch_lab/karamanlis_2024/sessions"
        stim_type (str): The stimulus type to filter files.
        normalize_stimuli (bool): Whether to normalize the stimuli.
        specie (str): Animal species (e.g., "mouse", "marmoset").
        downsampled_size (tuple[int, int]): Size to downsample the stimuli.

    Returns:
        dict[str, MoviesTrainTestSplit]: Dictionary mapping session names to stimulus data.
    """
    base_data_path = get_local_file_path(str(base_data_path))

    stimuli_all_sessions = {}
    exp_sessions = [
        path
        for path in os.listdir(base_data_path)
        if (os.path.isdir(os.path.join(base_data_path, path)) or path.endswith(".zip")) and specie in path
    ]

    assert len(exp_sessions) > 0, (
        f"No data directories found in {base_data_path} for animal {specie}."
        "Please check the path and the animal species provided, and that you unrared the data files."
    )

    for session in tqdm(exp_sessions, desc="Processing sessions"):
        session_path = os.path.join(base_data_path, session)
        stimuli = load_stimuli_for_session(session_path, stim_type, downsampled_size, normalize_stimuli)
        if stimuli:
            stimuli_all_sessions[session.split("/")[-1]] = stimuli

    return stimuli_all_sessions


def return_fix_movie_torch(
    screensize: tuple[int, int], im_ensemble: np.ndarray | torch.Tensor, list_fixations: np.ndarray
) -> torch.Tensor:
    """
    Generates a movie of fixations using an image ensemble.
    Faster version using PyTorch, which supports fancier indexing.

    Python port from: https://github.com/dimokaramanlis/subunit_grid_model/blob/main/code/returnFixMovie.m

    Parameters:
    screensize : tuple
        (Ny, Nx) size of the monitor, e.g., (600, 800).
    im_ensemble : ndarray
        Ensemble of images with dimensions (Nyim, Nxim, Nimages).
    list_fixations : ndarray
        Array of shape (3, Nframes) containing fixations (image index, x, y).

    Returns:
    blockstimulus : ndarray
        Movie of fixations with shape (Ny, Nx, Nframes), where the images
        presented at each frame are shifted based on gaze data.
    """
    Ny, Nx = screensize
    Nyim, Nxim, Nimages = im_ensemble.shape
    Nframes = list_fixations.shape[1]

    blockstimulus = torch.zeros((Ny, Nx, Nframes), dtype=torch.float32)
    im_ensemble = torch.tensor(im_ensemble, dtype=torch.float32)
    list_fixations_tensor = torch.tensor(list_fixations, dtype=torch.int64)

    for ifix in range(Nframes):
        xmin, xmax, ymin, ymax, x_or, y_or = get_ranges(
            list_fixations[1, ifix],
            list_fixations[2, ifix],
            Nx,
            Ny,
            Nxim,
            Nyim,
            np.arange(1, Nx + 1),
            np.arange(1, Ny + 1),
        )
        x_or = torch.tensor(x_or, dtype=torch.int64).unsqueeze(0)  # type: ignore
        y_or = torch.tensor(y_or, dtype=torch.int64).unsqueeze(1)  # type: ignore
        blockstimulus[ymin : ymax + 1, xmin : xmax + 1, ifix] = im_ensemble[
            y_or, x_or, list_fixations_tensor[0, ifix] - 1
        ]

    return blockstimulus


def return_fix_movie(screensize: tuple[int, int], im_ensemble: np.ndarray, list_fixations: np.ndarray) -> np.ndarray:
    """
    Generates a movie of fixations using an image ensemble.

    Python port from: https://github.com/dimokaramanlis/subunit_grid_model/blob/main/code/returnFixMovie.m

    Parameters:
    screensize : tuple
        (Ny, Nx) size of the monitor, e.g., (600, 800).
    im_ensemble : ndarray
        Ensemble of images with dimensions (Nyim, Nxim, Nimages).
    list_fixations : ndarray
        Array of shape (3, Nframes) containing fixations (image index, x, y).

    Returns:
    blockstimulus : ndarray
        Movie of fixations with shape (Ny, Nx, Nframes), where the images
        presented at each frame are shifted based on gaze data.
    """
    Ny, Nx = screensize
    Nyim, Nxim, Nimages = im_ensemble.shape
    Nframes = list_fixations.shape[1]

    blockstimulus = np.zeros((Ny, Nx, Nframes), dtype=np.float32)

    for ifix in range(Nframes):
        xmin, xmax, ymin, ymax, x_or, y_or = get_ranges(
            list_fixations[1, ifix],
            list_fixations[2, ifix],
            Nx,
            Ny,
            Nxim,
            Nyim,
            np.arange(1, Nx + 1),
            np.arange(1, Ny + 1),
        )
        x_or = x_or.astype(int)
        y_or = y_or.astype(int)
        blockstimulus[ymin : ymax + 1, xmin : xmax + 1, ifix] = im_ensemble[np.ix_(y_or, x_or)][
            :, :, list_fixations[0, ifix] - 1
        ]

    return blockstimulus


def get_ranges(
    tr_x: int, tr_y: int, Nxs: int, Nys: int, Nx: int, Ny: int, rx: np.ndarray, ry: np.ndarray
) -> tuple[int, int, int, int, np.ndarray, np.ndarray]:
    """
    Compute valid ranges and indices based on translations.

    Parameters:
    tr_x, tr_y : int
        Translation values for x and y.
    Nxs, Nys : int
        Screen dimensions (Nx, Ny).
    Nx, Ny : int
        Image dimensions.
    rx, ry : ndarray
        Ranges for x and y.

    Returns:
    xmin, xmax, ymin, ymax : int
        Indices for cropping (screen coordinates).
    x_or, y_or : ndarray
        Translated indices for image coordinates.
    """
    ymin = max((Nys // 2) - tr_y, 0)  # Minimum y index for the screen
    ymax = min(Ny + (Nys // 2) - tr_y, Nys)  # Maximum y index for the screen

    xmin = max((Nxs // 2) - tr_x, 0)  # Minimum x index for the screen
    xmax = min(Nx + (Nxs // 2) - tr_x, Nxs)  # Maximum x index for the screen

    # Filter valid indices within these bounds
    rx_use = (rx >= xmin) & (rx <= xmax)
    ry_use = (ry >= ymin) & (ry <= ymax)

    # Translate these to the original image coordinates
    x_or = tr_x - (Nxs // 2) + rx[rx_use] - 1
    y_or = tr_y - (Nys // 2) + ry[ry_use] - 1

    # Find the bounds again for the block on the screen
    xmin = np.where(rx_use)[0][0]
    xmax = np.where(rx_use)[0][-1]
    ymin = np.where(ry_use)[0][0]
    ymax = np.where(ry_use)[0][-1]

    return xmin, xmax, ymin, ymax, x_or, y_or



"""
Minimal responses loading utilities to train a model on the data used in Karamanlis et al. 2024

Paper: https://doi.org/10.1038/s41586-024-08212-3
Data: https://doi.org/10.12751/g-node.ejk8kx
"""

import os
import warnings
from pathlib import Path
from typing import Literal

import numpy as np
from einops import rearrange
from tqdm.auto import tqdm

from openretina.data_io.base import ResponsesTrainTestSplit
from openretina.utils.file_utils import get_local_file_path, unzip_and_cleanup
from openretina.utils.h5_handling import load_dataset_from_h5


def load_responses_for_session(
    session_path: str | os.PathLike,
    stim_type: Literal["fixationmovie", "frozencheckerflicker", "gratingflicker", "imagesequence"],
    fr_normalisation: float,
) -> ResponsesTrainTestSplit | None:
    """
    Load responses for a single session.

    Args:
        session_path (str): Path to the session directory.
        stim_type (str): The stimulus type to filter files.
        fr_normalisation (float): Normalization factor for firing rates.

    Returns:
        ResponsesTrainTestSplit: Loaded responses for the session.

    Raises:
        IOError: If multiple relevant files are found.
    """
    if str(session_path).endswith(".zip"):
        session_path = unzip_and_cleanup(Path(session_path))
    recording_file_names = [x for x in os.listdir(session_path) if x.endswith(f"{stim_type}_data.mat")]

    if len(recording_file_names) == 0:
        warnings.warn(
            f"No file with postfix {f'{stim_type}_data.mat'} found in folder {session_path}", UserWarning, stacklevel=2
        )
        return None
    elif len(recording_file_names) > 1:
        raise IOError(
            f"Multiple files with postfix {f'{stim_type}_data.mat'} found in folder {session_path}: "
            f"{recording_file_names=}"
        )

    full_path = os.path.join(session_path, recording_file_names[0])
    tqdm.write(f"Loading responses from {full_path}")

    testing_responses = load_dataset_from_h5(full_path, "frozenbin")
    training_responses = load_dataset_from_h5(full_path, "runningbin")

    testing_responses = rearrange(testing_responses, "trials time neurons -> trials neurons time") / fr_normalisation
    mean_test_response = np.mean(testing_responses, axis=0)

    training_responses = rearrange(training_responses, "block time neurons -> neurons (block time)") / fr_normalisation

    return ResponsesTrainTestSplit(
        train=training_responses,
        test=mean_test_response,
        test_by_trial=testing_responses,
        stim_id=stim_type,
    )


def load_all_responses(
    base_data_path: str | os.PathLike,
    stim_type: Literal["fixationmovie", "frozencheckerflicker", "gratingflicker", "imagesequence"] = "fixationmovie",
    specie: Literal["mouse", "marmoset"] = "mouse",
    fr_normalization: float = 1.0,
) -> dict[str, ResponsesTrainTestSplit]:
    """
    Load responses for all sessions.

    Args:
        base_data_path (str | os.PathLike): Base directory containing session data.
                                            Can also be the path to the "sessions" folder in the huggingface mirror.
        "https://huggingface.co/datasets/open-retina/open-retina/tree/main/gollisch_lab/karamanlis_2024/sessions"
        stim_type (str): The stimulus type to filter files.
        specie (str): Animal species (e.g., "mouse", "marmoset").
        fr_normalization (float): Normalization factor for firing rates.

    Returns:
        dict[str, ResponsesTrainTestSplit]: Dictionary mapping session names to response data.
    """
    base_data_path = get_local_file_path(str(base_data_path))

    responses_all_sessions = {}
    exp_sessions = [
        path
        for path in os.listdir(base_data_path)
        if (os.path.isdir(os.path.join(base_data_path, path)) or path.endswith(".zip")) and specie in path
    ]

    assert len(exp_sessions) > 0, (
        f"No data directories found in {base_data_path} for animal {specie}."
        "Please check the path and the animal species provided, and that you unrared the data files."
    )

    for session in tqdm(exp_sessions, desc="Processing sessions"):
        session_path = os.path.join(base_data_path, session)
        responses = load_responses_for_session(session_path, stim_type, fr_normalization)
        if responses:
            responses_all_sessions[str(os.path.basename(os.path.normpath(session)))] = responses

    return responses_all_sessions



import torch
from openretina.models import *
# Download model
model = load_core_readout_from_remote("hoefling_2024_base_low_res", "cpu")
# Run forward pass
responses = model.forward(torch.rand(model.stimulus_shape(time_steps=50)))



import os
from typing import Literal

import torch
import torch.nn as nn

from openretina.modules.readout.base import ClonedReadout, Readout
from openretina.modules.readout.factorised_gaussian import SimpleSpatialXFeature3d


class MultiGaussianReadoutWrapper(nn.ModuleDict):
    """
    Multiple Sessions version of the SimpleSpatialXFeature3d factorised gaussian readout.
    """

    def __init__(
        self,
        in_shape: tuple[int, int, int, int],
        n_neurons_dict: dict[str, int],
        scale: bool,
        bias: bool,
        gaussian_masks: bool,
        gaussian_mean_scale: float,
        gaussian_var_scale: float,
        positive: bool,
        gamma_readout: float,
        gamma_masks: float = 0.0,
        readout_reg_avg: bool = False,
    ):
        super().__init__()
        for k in n_neurons_dict:  # iterate over sessions
            n_neurons = n_neurons_dict[k]
            assert len(in_shape) == 4
            self.add_module(
                k,
                SimpleSpatialXFeature3d(  # add a readout for each session
                    in_shape,
                    n_neurons,
                    gaussian_mean_scale=gaussian_mean_scale,
                    gaussian_var_scale=gaussian_var_scale,
                    positive=positive,
                    scale=scale,
                    bias=bias,
                ),
            )

        self.gamma_readout = gamma_readout
        self.gamma_masks = gamma_masks
        self.gaussian_masks = gaussian_masks
        self.readout_reg_avg = readout_reg_avg

    def forward(self, *args, data_key: str | None, **kwargs) -> torch.Tensor:
        if data_key is None:
            readout_responses = []
            for readout_key in self.readout_keys():
                resp = self[readout_key](*args, **kwargs)
                readout_responses.append(resp)
            response = torch.concatenate(readout_responses, dim=-1)
        else:
            response = self[data_key](*args, **kwargs)
        return response

    def regularizer(self, data_key: str) -> torch.Tensor:
        feature_loss = self[data_key].feature_l1(average=self.readout_reg_avg) * self.gamma_readout
        mask_loss = self[data_key].mask_l1(average=self.readout_reg_avg) * self.gamma_masks
        return feature_loss + mask_loss

    def readout_keys(self) -> list[str]:
        return sorted(self._modules.keys())

    def save_weight_visualizations(self, folder_path: str) -> None:
        for key in self.readout_keys():
            readout_folder = os.path.join(folder_path, key)
            os.makedirs(readout_folder, exist_ok=True)
            self._modules[key].save_weight_visualizations(readout_folder)  # type: ignore

    @property
    def sessions(self) -> list[str]:
        return self.readout_keys()


class MultiReadoutBase(nn.ModuleDict):
    """
    Base class for MultiReadouts. It is a dictionary of data keys and readouts to the corresponding datasets.

    Adapted from neuralpredictors. Original code at:
    https://github.com/sinzlab/neuralpredictors/blob/v0.3.0.pre/neuralpredictors/layers/readouts/multi_readout.py


    Args:
        in_shape_dict (dict): dictionary of data_key and the corresponding dataset's shape as an output of the core.

        n_neurons_dict (dict): dictionary of data_key and the corresponding dataset's number of neurons

        base_readout (torch.nn.Module): base readout class. If None, self._base_readout must be set manually in the
                                        inheriting class's definition.

        mean_activity_dict (dict): dictionary of data_key and the corresponding dataset's mean responses.
                                    Used to initialize the readout bias with.
                                    If None, the bias is initialized with 0.

        clone_readout (bool): whether to clone the first data_key's readout to all other readouts, only allowing for a
                                scale and offset. This is a rather simple method to enforce parameter-sharing
                                between readouts.

        gamma_readout (float): regularization strength

        **kwargs: additional keyword arguments to be passed to the base_readout's constructor
    """

    _base_readout = None

    def __init__(
        self,
        in_shape: tuple[int, int, int, int],
        n_neurons_dict: dict[str, int],
        base_readout: Readout | None = None,
        mean_activity_dict: dict[str, float] | None = None,
        clone_readout=False,
        **kwargs,
    ):
        # The `base_readout` can be overridden only if the static property `_base_readout` is not set
        if self._base_readout is None:
            self._base_readout = base_readout

        if self._base_readout is None:
            raise ValueError("Attribute _base_readout must be set")
        super().__init__()

        for i, data_key in enumerate(n_neurons_dict):
            mean_activity = mean_activity_dict[data_key] if mean_activity_dict is not None else None

            if i == 0 or clone_readout is False:
                self.add_module(
                    data_key,
                    self._base_readout(
                        in_shape=in_shape,
                        outdims=n_neurons_dict[data_key],
                        mean_activity=mean_activity,
                        **kwargs,
                    ),
                )
                original_readout = data_key
            elif i > 0 and clone_readout is True:
                self.add_module(data_key, ClonedReadout(self[original_readout]))

        self.initialize(mean_activity_dict)

    def forward(self, *args, data_key: str | None = None, **kwargs):
        if data_key is None and len(self) == 1:
            data_key = list(self.keys())[0]
        return self[data_key](*args, **kwargs)

    def initialize(self, mean_activity_dict: dict[str, float] | None = None):
        for data_key, readout in self.items():
            mean_activity = mean_activity_dict[data_key] if mean_activity_dict is not None else None
            readout.initialize(mean_activity)

    def regularizer(self, data_key: str | None = None, reduction: Literal["sum", "mean", None] = "sum"):
        if data_key is None and len(self) == 1:
            data_key = list(self.keys())[0]
        return self[data_key].regularizer(reduction=reduction)



import pytest
import torch

from openretina.utils.misc import check_server_responding
from openretina.utils.nnfabrik_model_loading import Center, load_ensemble_model_from_remote


@pytest.mark.parametrize(
    "session_id, center_readout",
    [
        ("2_ventral2_20201016", None),
        ("1_ventral1_20201021", None),
        ("2_ventral1_20201021", Center(target_mean=(0.0, 0.0))),
    ],
)
@pytest.mark.skipif(
    condition=not check_server_responding("https://gin.g-node.org/eulerlab/rgc-natstim/"),
    reason="GIN server unreachable.",
)
def test_loading_model_from_remote(session_id: str, center_readout: Center | None) -> None:
    data_info, ensemble_model = load_ensemble_model_from_remote(device="cpu", center_readout=center_readout)
    assert session_id in data_info

    # run a forward pass with a zero tensor of batch size 1
    input_dim = data_info[session_id]["input_dimensions"]
    input_dim_batch_size_one = (1,) + input_dim[1:]
    inp_ = torch.zeros(input_dim_batch_size_one)
    out = ensemble_model(inp_, session_id)

    expected_output_dim = (1, input_dim[2] - 30, data_info[session_id]["output_dimension"])
    assert out.shape == expected_output_dim



from copy import deepcopy

import numpy as np
import torch
from torch import nn, optim

from openretina.insilico.stimulus_optimization.objective import IncreaseObjective


class MeiAcrossContrasts(nn.Module):
    """
    PyTorch Module wrapper around a stimulus to allow tracking response gradients onto the stimulus.
    """

    def __init__(self, contrast_values: torch.Tensor, stim: torch.Tensor):
        """
        Parameters:
        contrast_values (torch.Tensor): A torch Tensor with 2 scalar values that the MEI green and UV channels will be
        multiplied with.
        stim (torch.Tensor): MEI stimulus as torch Tensor.
        """
        super().__init__()
        self.contrast_values = nn.Parameter(contrast_values, requires_grad=True)
        self.stim = nn.Parameter(stim, requires_grad=False)

    def forward(self):
        """
        Multiplies each channel of the stimulus with the contrast value in the corresponding channel.
        This yields the stimulus at the location in the contrast grid specified by the contrast values.
        """
        return torch.mul(
            torch.stack(
                [
                    torch.ones_like(self.stim[:, 0, ...]) * self.contrast_values[0],
                    torch.ones_like(self.stim[:, 0, ...]) * self.contrast_values[1],
                ],
                dim=1,
            ),
            self.stim.squeeze(),
        )


def trainer_fn(
    mei_contrast_gen: MeiAcrossContrasts,
    model_neuron: IncreaseObjective,
    optimizer_class=optim.Adam,
    lr: float = 1.0,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Trainer function for getting the gradient on the MEI with different contrasts
    """
    optimizer = optimizer_class(mei_contrast_gen.parameters(), lr=lr)
    loss = model_neuron.forward(mei_contrast_gen())
    loss.backward()
    grad_val: torch.Tensor = deepcopy(mei_contrast_gen.contrast_values.grad)  # type: ignore
    optimizer.zero_grad()
    return grad_val.detach().cpu().numpy().squeeze(), loss.detach().cpu().numpy().squeeze()


def get_gradient_grid(
    stim: torch.Tensor,
    model_neuron: IncreaseObjective,
    n_channels: int = 2,
    start: float = -1,
    stop: float = 1,
    step_size: float = 0.1,
) -> tuple:
    """
    Generate a grid of response gradients for a given stimulus and model neuron.

    Args:
        stim (torchTensor, 1 x n_channels x time x height x width): The MEI stimulus.
        model_neuron (IncreaseObjective): The model neuron objective.
        n_channels (int, optional): The number of channels. Defaults to 2.
        start (float, optional): The starting value for the contrast range. Defaults to -1.
        stop (float, optional): The ending value for the contrast range. Defaults to 1.
        step_size (float, optional): The step size for the contrast range. Defaults to 0.1.

    Returns:
        tuple: A tuple containing the following elements:
            - grid (ndarray): A grid of gradient values with shape (n_channels, len(green_contrast_values),
            len(uv_contrast_values)).
            - resp_grid (ndarray): A grid of loss values with shape (len(green_contrast_values),
            len(uv_contrast_values)).
            - norm_grid (ndarray): A grid of norm values with shape (len(green_contrast_values),
            len(uv_contrast_values)).
            - green_contrast_values (ndarray): An array of green contrast values.
            - uv_contrast_values (ndarray): An array of UV contrast values.
    """
    green_contrast_values = np.arange(start, stop + step_size, step_size)
    uv_contrast_values = np.arange(start, stop + step_size, step_size)
    grid = np.zeros((n_channels, len(green_contrast_values), len(uv_contrast_values)))
    resp_grid = np.zeros((len(green_contrast_values), len(uv_contrast_values)))
    norm_grid = np.zeros((len(green_contrast_values), len(uv_contrast_values)))

    for i, contrast_green in enumerate(np.arange(-1, 1 + step_size, step_size)):
        for j, contrast_uv in enumerate(np.arange(-1, 1 + step_size, step_size)):
            mei_contrast_gen = MeiAcrossContrasts(torch.Tensor([contrast_green, contrast_uv]), stim)
            response_gradient, response = trainer_fn(mei_contrast_gen, model_neuron, lr=0.1)
            grid[0, i, j] = response_gradient[0]
            grid[1, i, j] = response_gradient[1]
            resp_grid[i, j] = response
            norm_grid[i, j] = np.linalg.norm(grid[:, i, j])
    return grid, resp_grid, norm_grid, green_contrast_values, uv_contrast_values


def equalize_channels(stim: torch.Tensor, flip_green: bool = False) -> torch.Tensor:
    """
    Scales the (green and UV) channels of an stim such that they have equal norm,
    and the scaled stim has the same norm as the original stim. Optionally flips sign of green channel
    :param stim: torch.Tensor, shape (1, 2, 50, 18, 16)
    :param flip_green: bool
    :return: equalized_stim: torch.Tensor, shape (1, 2, 50, 18, 16)
    """
    green_chan = stim[0, 0]
    uv_chan = stim[0, 1]
    green_norm = torch.norm(green_chan)
    uv_norm = torch.norm(uv_chan)
    total_norm = torch.norm(stim)
    green_factor = (total_norm / 2) / green_norm
    uv_factor = (total_norm / 2) / uv_norm
    equalized_stim = torch.zeros_like(stim, device=stim.device)
    equalized_stim[0, 0] = green_factor * stim[0, 0]
    if flip_green:
        equalized_stim[0, 0] = -1 * equalized_stim[0, 0]
    equalized_stim[0, 1] = uv_factor * stim[0, 1]
    total_factor = total_norm / torch.norm(equalized_stim)
    equalized_stim = total_factor * equalized_stim
    return equalized_stim



from openretina.utils.file_utils import get_local_file_path
from openretina.utils.plotting import plot_stimulus_composition, save_stimulus_to_mp4_video

__all__ = ["get_local_file_path", "save_stimulus_to_mp4_video", "plot_stimulus_composition"]



from openretina.insilico.stimulus_optimization.objective import (
    ContrastiveNeuronObjective,
    IncreaseObjective,
    MeanReducer,
    SliceMeanReducer,
)
from openretina.insilico.stimulus_optimization.optimization_stopper import OptimizationStopper
from openretina.insilico.stimulus_optimization.optimizer import optimize_stimulus

__all__ = [
    "optimize_stimulus",
    "IncreaseObjective",
    "ContrastiveNeuronObjective",
    "OptimizationStopper",
    "MeanReducer",
    "SliceMeanReducer",
]
